'''
Defines a tensorial field of rank 2 over a grid

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

from typing import Tuple, Optional, Union, Callable, TYPE_CHECKING

import numpy as np
import numba as nb

from .base import DataFieldBase
from .scalar import ScalarField
from .vectorial import VectorField

from ..tools.numba import jit
from ..tools.docstrings import fill_in_docstring



if TYPE_CHECKING:
    from ..grids.boundaries.axes import BoundariesData  # @UnusedImport
    
    
    
class Tensor2Field(DataFieldBase):
    """ Single tensor field of rank 2 on a grid
    
    Attributes:
        grid (:class:`~pde.grids.GridBase`):
            The underlying grid defining the discretization
        data (:class:`numpy.ndarray`):
            Tensor components at the support points of the grid
        label (str):
            Name of the field
    """

    rank = 2
         
          
    def __getitem__(self, key: Tuple[int, int]) -> ScalarField:
        """ extract a component of the VectorField """
        try:
            if not (isinstance(key[0], int) and isinstance(key[1], int)):
                raise IndexError('Index must be given as two integers')
        except TypeError:
            raise IndexError('Index must be given as two integers')
        return ScalarField(self.grid, self.data[key]) 
                           
     
    @DataFieldBase._data_flat.setter  # type: ignore
    def _data_flat(self, value):
        """ set the data from a value from a collection """
        dim = self.grid.dim
        self._data = value.reshape(dim, dim, *self.grid.shape)
        # check whether both point to the same memory location
        assert (value.__array_interface__['data'][0] ==
                self._data.__array_interface__['data'][0])  
                
     
    def dot(self, other: Union[VectorField, "Tensor2Field"],
            out: Optional[Union[VectorField, "Tensor2Field"]] = None,
            label: str = 'dot product') -> Union[VectorField, "Tensor2Field"]:
        """ calculate the dot product involving a tensor field
        
        This supports the dot product between two tensor fields as well as the
        product between a tensor and a vector. The resulting fields will be a
        tensor or vector, respectively.
        
        Args:
            other (VectorField or Tensor2Field):
                the second field
            out (VectorField or Tensor2Field, optional):
                Optional field to which the  result is written.
            label (str, optional):
                Name of the returned field
            
        Returns:
            VectorField or Tensor2Field: the result of applying the dot operator             
        """
        # check input
        self.grid.assert_grid_compatible(other.grid)
        if not isinstance(other, (VectorField, Tensor2Field)):
            raise TypeError('Second term must be a vector or tensor field')

        # create and check the output instance
        if out is None:
            out = other.__class__(self.grid)
        else:
            assert isinstance(out, other.__class__)
            self.grid.assert_grid_compatible(out.grid)
            
        # calculate the result
        np.einsum('ij...,j...->i...', self.data, other.data, out=out.data)
        if label is not None:
            out.label = label
            
        return out
    
    __matmul__ = dot  # support python @-syntax for matrix multiplication
    
    
    def get_dot_operator(self) -> Callable:
        """ return operator calculating the dot product involving vector fields
        
        This supports both products between two vectors as well as products
        between a vector and a tensor.
        
        Warning:
            This function does not check types or dimensions.
        
        Returns:
            function that takes two instance of :class:`numpy.ndarray`, which
            contain the discretized data of the two operands. An optional third
            argument can specify the output array to which the result is
            written. Note that the returned function is jitted with numba for
            speed.
        """
        dim = self.grid.dim
                
        @jit
        def inner(a, b, out):
            """ calculate dot product between fields `a` and `b` """
            for i in range(dim):
                out[i] = a[i, 0] * b[0]  # overwrite potential data in out
                for j in range(1, dim):
                    out[i] += a[i, j] * b[j]
            return out
        
        if nb.config.DISABLE_JIT:
            def dot(a, b, out=None):
                """ wrapper deciding whether the underlying function is called
                with or without `out`. """
                if out is None:
                    out = np.empty_like(b)
                return inner(a, b, out)
            
        else:
            @nb.generated_jit
            def dot(a, b, out=None):
                """ wrapper deciding whether the underlying function is called
                with or without `out`. """
                if isinstance(a, nb.types.Number):
                    # simple scalar call -> do not need to allocate anything
                    raise RuntimeError('Dot needs to be called with fields')
                
                elif isinstance(out, (nb.types.NoneType, nb.types.Omitted)):
                    # function is called without `out`
                    def f_with_allocated_out(a, b, out):
                        """ helper function allocating output array """
                        return inner(a, b, out=np.empty_like(b))
                
                    return f_with_allocated_out
                
                else:
                    # function is called with `out` argument
                    return inner          
        
        return dot
    
          
    @fill_in_docstring
    def divergence(self, bc: "BoundariesData",
                   out: Optional[VectorField] = None,
                   label: str = 'divergence') -> VectorField:
        """ apply (tensor) divergence and return result as a field 
        
        Args:
            bc: 
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES}
            out (VectorField, optional):
                Optional scalar field to which the  result is written.
            label (str, optional):
                Name of the returned field
            
        Returns:
            VectorField: the result of applying the operator 
        """
        tensor_divergence = self.grid.get_operator('tensor_divergence', bc=bc)
        if out is None:
            out = VectorField(self.grid, tensor_divergence(self.data),
                              label=label)
        else:
            assert isinstance(out, VectorField)
            tensor_divergence(self.data, out=out.data)
        return out
    
    
    @property
    def integral(self):
        """ :class:`numpy.ndarray`: integral of each component over space """
        return self.grid.integrate(self.data)
    
    
    def transpose(self, label: str = 'transpose') -> 'Tensor2Field':
        """ return the transpose of the tensor field
        
        Args:
            label (str, optional): Name of the returned field
            
        Returns:
            Tensor2Field: holding the transpose of the tensor field      
        """
        axes = (1, 0) + tuple(range(2, 2 + self.grid.dim))
        return Tensor2Field(self.grid, self.data.transpose(axes), label=label)
    
    
    def symmetrize(self, make_traceless: bool = False, inplace: bool = False) \
            -> "Tensor2Field":
        """ symmetrize the tensor field in place
        
        Args:
            make_traceless (bool):
                Determines whether the result is also traceless
            inplace (bool):
                Flag determining whether to symmetrize the current field or
                return a new one 
        """
        if inplace:
            out = self
        else:
            out = self.copy()
        
        out += self.transpose()  # type: ignore
        out *= 0.5  # type: ignore
        
        if make_traceless:
            dim = self.grid.dim
            value = self.trace() / dim
            for i in range(dim):
                out._data[i, i] -= value.data
        return out
    
    
    def to_scalar(self, scalar: str = 'auto',
                  label: Optional[str] = 'scalar `{scalar}`') -> ScalarField:
        r""" return a scalar field by applying `method`
        
        The invariants of the tensor field :math:`\boldsymbol{A}` are
        
        .. math::
            I_1 &= \mathrm{tr}(\boldsymbol{A}) \\
            I_2 &= \frac12 \left[
                (\mathrm{tr}(\boldsymbol{A})^2 -
                \mathrm{tr}(\boldsymbol{A}^2)
            \right] \\
            I_3 &= \det(A)
            
        where `tr` denotes the trace and `det` denotes the determinant.
        Note that the three invariants can only be distinct and non-zero in 
        three dimensions. In two dimensional spaces, we have the identity
        :math:`2 I_2 = I_3` and in one-dimensional spaces, we have
        :math:`I_1 = I_3` as well as :math:`I_2 = 0`.
            
        Args:
            scalar (str):
                Choose the method to use. Possible choices include `norm` (the 
                default), `min`, `max`, `squared_sum`, `trace` (or
                `invariant1`), `invariant2`, and `determinant` (or `invariant3`)
            label (str, optional):
                Name of the returned field
            
        Returns:
            :class:`pde.fields.scalar.ScalarField`: the scalar field after
            applying the operation
        """
        if scalar == 'auto':
            scalar = 'norm'
        
        if scalar == 'norm':
            data = np.linalg.norm(self.data, axis=(0, 1))
            
        elif scalar == 'min':
            data = np.min(self.data, axis=(0, 1))
            
        elif scalar == 'max':
            data = np.max(self.data, axis=(0, 1))
            
        elif scalar == 'squared_sum':
            data = np.sum(self.data**2, axis=(0, 1))
            
        elif scalar == 'trace' or scalar == 'invariant1':
            data = self.data.trace(axis1=0, axis2=1)
            
        elif scalar == 'invariant2':
            data = np.zeros(self.grid.shape)
            for i in range(self.grid.dim):
                for j in range(i):
                    data += (self.data[i, i] * self.data[j, j] -
                             self.data[i, j] * self.data[j, i])
            data *= 0.5
            
        elif scalar in {'det', 'determinant', 'invariant3'}:
            if self.grid.dim == 1:
                data = self.data[0, 0]
            else:
                data = np.zeros(self.grid.shape)
                # this iterates over all of space and might thus be slow, but
                # the interface of np.linalg.det is not very flexible. We could
                # in principle use the definition of np.linalg.det without the
                # multiple checks to gain some speed
                for i in np.ndindex(self.grid.shape):
                    i2 = (Ellipsis, ) + i  # type: ignore
                    data[i] = np.linalg.det(self.data[i2])
                
        else:
            raise ValueError(f'Unknown method `{scalar}` for `to_scalar`. '
                             'Valid methods are `norm`, `min`, `max`, '
                             'squared_sum`, `trace`, `determinant`, and '
                             '`invariant#`, where # is 1, 2, or 3')

        # determine label of the result
        if self.label is None:
            if label is not None:
                label = label.format(scalar=scalar)
        else:
            label = f'{scalar} of {self.label}'
            
        return ScalarField(self.grid, data, label=label)
    
    
    def trace(self, label: Optional[str] = 'trace') -> ScalarField:
        """ return the trace of the tensor field as a scalar field
        
        Args:
            label (str, optional): Name of the returned field
            
        Returns:
            ScalarField: holding the trace
        """
        return self.to_scalar(scalar='trace', label=label)

    



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import numpy as np
import pytest

from ..local import _get_arr_1d, BCBase
from ... import UnitGrid



def test_get_arr_1d():
    """ test the _get_arr_1d function """
    # 1d
    a = np.arange(3)
    arr_1d, i, bc_idx = _get_arr_1d(a, [1], 0)
    assert i == 1
    assert bc_idx == (...,)
    np.testing.assert_equal(arr_1d, a)

    # 2d
    a = np.arange(4).reshape(2, 2)
    arr_1d, i, bc_idx = _get_arr_1d(a, [0, 0], 0)
    assert i == 0
    assert bc_idx == (..., 0)
    np.testing.assert_equal(arr_1d, a[:, 0])
    
    arr_1d, i, bc_idx = _get_arr_1d(a, [1, 1], 1)
    assert i == 1
    assert bc_idx == (..., 1)
    np.testing.assert_equal(arr_1d, a[1, :])

    # 3d
    a = np.arange(8).reshape(2, 2, 2)
    arr_1d, i, bc_idx = _get_arr_1d(a, [0, 0, 0], 0)
    assert i == 0
    assert bc_idx == (..., 0, 0)
    np.testing.assert_equal(arr_1d, a[:, 0, 0])
    
    arr_1d, i, bc_idx = _get_arr_1d(a, [1, 1, 0], 1)
    assert i == 1
    assert bc_idx == (..., 1, 0)
    np.testing.assert_equal(arr_1d, a[1, :, 0])
    
    arr_1d, i, bc_idx = _get_arr_1d(a, [1, 1, 0], 2)
    assert i == 0
    assert bc_idx == (..., 1, 1)
    np.testing.assert_equal(arr_1d, a[1, 1, :])



def test_individual_boundaries():
    """ test setting individual boundaries """
    g = UnitGrid([2])
    for data in ['value', {'value': 1}, {'type': 'value', 'value': 1},
                 'natural', 'derivative', {'derivative': 1},
                 {'type': 'derivative', 'value': 1}, {'mixed': 1},
                 {'type': 'mixed', 'value': 1}, 'extrapolate']:
        bc = BCBase.from_data(g, 0, upper=True, data=data, rank=0)
        
        assert isinstance(str(bc), str)
        assert isinstance(repr(bc), str)
        assert bc.rank == 0
        assert bc.homogeneous
        bc.check_value_rank(0)
        with pytest.raises(RuntimeError):
            bc.check_value_rank(1)

        for bc_copy in [BCBase.from_data(g, 0, upper=True, data=bc, rank=0),
                        bc.copy()]:
            assert bc == bc_copy
            assert bc._cache_hash() == bc_copy._cache_hash()

    assert bc.extract_component() == bc
        


def test_individual_boundaries_multidimensional():
    """ test setting individual boundaries in 2d """
    g2 = UnitGrid([2, 3])
    bc = BCBase.from_data(g2, 0, True, {'type': 'value', 'value': [1, 2]},
                          rank=1)

    assert isinstance(str(bc), str)
    assert isinstance(repr(bc), str)
    assert bc.rank == 1
    assert bc.homogeneous
    bc.check_value_rank(1)
    with pytest.raises(RuntimeError):
        bc.check_value_rank(0)
        
    bc_comp = bc.extract_component(0)
    assert bc_comp.rank == 0
    assert bc_comp.value == 1
    bc_comp = bc.extract_component(1)
    assert bc_comp.rank == 0
    assert bc_comp.value == 2
    
    for bc_copy in [BCBase.from_data(g2, 0, upper=True, data=bc, rank=1),
                    bc.copy()]:
        assert bc == bc_copy
        assert bc._cache_hash() == bc_copy._cache_hash()

    
        
def test_virtual_points():
    """ test the calculation of virtual points """
    g = UnitGrid([2])
    data = np.array([1, 2])
    
    # test constant boundary conditions
    bc = BCBase.from_data(g, 0, False, {'type': 'value', 'value': 1})
    assert bc.get_virtual_point(data) == pytest.approx(1)
    assert not bc.value_is_linked
    bc = BCBase.from_data(g, 0, True, {'type': 'value', 'value': 1})
    assert bc.get_virtual_point(data) == pytest.approx(0)
    assert not bc.value_is_linked
    
    # test derivative boundary conditions (wrt to outwards derivative)
    for up, b, val in [(False, {'type': 'derivative', 'value': -1}, 0),
                        (True, {'type': 'derivative', 'value': 1}, 3),
                        (False, 'extrapolate', 0),
                        (True, 'extrapolate', 3),
                        (False, {'type': 'mixed', 'value': 4, 'const': 1}, 0),
                        (True, {'type': 'mixed', 'value': 2, 'const': 4}, 2)]:
        bc = BCBase.from_data(g, 0, up, b)
        assert bc.upper == up
        assert bc.get_virtual_point(data) == pytest.approx(val)
        assert not bc.value_is_linked
        ev = bc.make_virtual_point_evaluator()
        assert ev(data, (2,) if up else (-1,)) == pytest.approx(val)
    
    # test curvature for y = 4 * x**2
    data = np.array([1, 9])
    bc = BCBase.from_data(g, 0, False, {'type': 'curvature', 'value': 8})
    assert bc.get_virtual_point(data) == pytest.approx(1)
    assert not bc.value_is_linked
    bc = BCBase.from_data(g, 0, True, {'type': 'curvature', 'value': 8})
    assert bc.get_virtual_point(data) == pytest.approx(25)
    assert not bc.value_is_linked

       
       
@pytest.mark.parametrize('upper', [False, True])
def test_virtual_points_linked_data(upper):
    """ test the calculation of virtual points with linked_data """
    g = UnitGrid([2, 2])
    point = (1, 1) if upper else (0, 0)
    data = np.zeros(g.shape)
    
    # test constant boundary conditions
    bc_data = np.array([1, 1])
    bc = BCBase.from_data(g, 0, upper, {'type': 'value', 'value': bc_data})
    assert not bc.value_is_linked
    bc.link_value(bc_data)
    assert bc.value_is_linked
    bc_data[:] = 3

    assert bc.get_virtual_point(data, point) == pytest.approx(6)
    ev = bc.make_virtual_point_evaluator()
    assert ev(data, point) == pytest.approx(6)

    # test derivative boundary conditions (wrt to outwards derivative)
    bc = BCBase.from_data(g, 0, upper, {'type': 'derivative', 'value': bc_data})
    assert not bc.value_is_linked
    bc.link_value(bc_data)
    assert bc.value_is_linked
    bc_data[:] = 4

    assert bc.get_virtual_point(data, point) == pytest.approx(4)
    ev = bc.make_virtual_point_evaluator()
    assert ev(data, point) == pytest.approx(4)
        
    # test derivative boundary conditions (wrt to outwards derivative)
    bc = BCBase.from_data(g, 0, upper, {'type': 'mixed', 'value': bc_data,
                                        'const': 3})
    assert not bc.value_is_linked
    bc.link_value(bc_data)
    assert bc.value_is_linked
    bc_data[:] = 4

    assert bc.get_virtual_point(data, point) == pytest.approx(1)
    ev = bc.make_virtual_point_evaluator()
    assert ev(data, point) == pytest.approx(1)
        
        
        
def test_mixed_condition():
    """ test the calculation of virtual points """
    g = UnitGrid([2])
    data = np.array([1, 2])
    
    for upper in [True, False]:
        bc1 = BCBase.from_data(g, 0, upper,
                               {'type': 'mixed', 'value': 0, 'const': 2})
        bc2 = BCBase.from_data(g, 0, upper, {'derivative': 2})
        assert bc1.get_virtual_point(data) == \
                pytest.approx(bc2.get_virtual_point(data))
                
    bc = BCBase.from_data(g, 0, False, {'type': 'mixed', 'value': np.inf})
    assert bc.get_virtual_point(data) == pytest.approx(-1)
    bc = BCBase.from_data(g, 0, True, {'type': 'mixed', 'value': np.inf})
    assert bc.get_virtual_point(data) == pytest.approx(-2)
    
    g = UnitGrid([2, 2])
    data = np.ones(g.shape)
    bc = BCBase.from_data(g, 0, False, {'type': 'mixed', 'value': [1, 2],
                                        'const': [3, 4]})
    assert bc.get_virtual_point(data, (0, 0)) == pytest.approx(2 + 1/3)
    assert bc.get_virtual_point(data, (0, 1)) == pytest.approx(2)
        
        
        
def test_inhomogeneous_bcs():
    """ test inhomogeneous boundary conditions """
    g = UnitGrid([2, 2])
    data = np.ones((2, 2))
    
    # first order bc
    bc_x = BCBase.from_data(g, 0, True, {'value': 'y'})
    assert isinstance(str(bc_x), str)
    assert bc_x.rank == 0
    assert bc_x.get_virtual_point(data, (1, 0)) == pytest.approx(0)
    assert bc_x.get_virtual_point(data, (1, 1)) == pytest.approx(2)

    # second order bc
    bc_x = BCBase.from_data(g, 0, True, {'curvature': 'y'})
    assert isinstance(str(bc_x), str)
    assert bc_x.rank == 0
    assert bc_x.get_virtual_point(data, (1, 0)) == pytest.approx(1.5)
    assert bc_x.get_virtual_point(data, (1, 1)) == pytest.approx(2.5)
    
    ev = bc_x.make_virtual_point_evaluator()
    assert ev(data, (1, 0)) == pytest.approx(1.5)
    assert ev(data, (1, 1)) == pytest.approx(2.5)

    ev = bc_x.make_adjacent_evaluator()
    assert ev(*_get_arr_1d(data, (0, 0), axis=0)) == pytest.approx(1)
    assert ev(*_get_arr_1d(data, (0, 1), axis=0)) == pytest.approx(1)
    assert ev(*_get_arr_1d(data, (1, 0), axis=0)) == pytest.approx(1.5)
    assert ev(*_get_arr_1d(data, (1, 1), axis=0)) == pytest.approx(2.5)
    # test lower bc
    bc_x = BCBase.from_data(g, 0, False, {'curvature': 'y'})
    ev = bc_x.make_adjacent_evaluator()
    assert ev(*_get_arr_1d(data, (1, 0), axis=0)) == pytest.approx(1)
    assert ev(*_get_arr_1d(data, (1, 1), axis=0)) == pytest.approx(1)
    assert ev(*_get_arr_1d(data, (0, 0), axis=0)) == pytest.approx(1.5)
    assert ev(*_get_arr_1d(data, (0, 1), axis=0)) == pytest.approx(2.5)



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import numpy as np

from ..numba import numba_environment, flat_idx, jit_allocate_out



def test_environment():
    """ test function signature checks """
    assert isinstance(numba_environment(), dict)
    


def test_flat_idx():
    """ test flat_idx function """
    assert flat_idx(2, 1) == 2
    assert flat_idx(np.arange(2), 1) == 1
    assert flat_idx(np.arange(4).reshape(2, 2), 1) == 1



def test_jit_allocate_out_1arg():
    """ test jit_allocate_out of functions with 1 argument """
    def f(arr, out):
        out[:] = arr
        return out
    
    a = np.linspace(0, 1, 3)
    g = jit_allocate_out(out_shape=a.shape)(f)
    np.testing.assert_equal(g(a), a)
    np.testing.assert_equal(jit_allocate_out(f)(a), a)
    


def test_jit_allocate_out_2arg():
    """ test jit_allocate_out of functions with 1 argument """
    def f(a, b, out):
        out[:] = a + b
        return out
    
    a = np.linspace(0, 1, 3)
    b = np.linspace(1, 2, 3)
    c = np.linspace(1, 3, 3)
    g = jit_allocate_out(out_shape=a.shape, num_args=2)(f)
    np.testing.assert_equal(g(a, b), c)
    np.testing.assert_equal(jit_allocate_out(num_args=2)(f)(a, b), c)
    


'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import os
import json

import pytest
import numpy as np

from .. import misc



def test_ensure_directory_exists(tmp_path):
    """ tests the ensure_directory_exists function """
    # create temporary name
    path = tmp_path / "test_ensure_directory_exists"
    assert not path.exists()
    # create the folder
    misc.ensure_directory_exists(path)
    assert path.is_dir()
    # check that a second call has the same result
    misc.ensure_directory_exists(path)
    assert path.is_dir()
    # remove the folder again
    os.rmdir(path)
    assert not path.exists()
    
    
    
def test_environment():
    """ test the environment function """
    assert isinstance(misc.environment(), dict)
    
    
    
def test_preserve_scalars():
    """ test the preserve_scalars decorator """
    class Test():
        @misc.preserve_scalars
        def meth(self, arr):
            return arr + 1
        
    t = Test()
    
    assert t.meth(1) == 2
    np.testing.assert_equal(t.meth(np.ones(2)), np.full(2, 2))
    
    
    
def test_hybridmethod():
    """ test the hybridmethod decorator """
    class Test():
        @misc.hybridmethod
        def meth(cls):  # @NoSelf
            return 'class'
        
        @meth.instancemethod
        def meth(self):
            return 'instance'
        
    assert Test.meth() == 'class'
    assert Test().meth() == 'instance'
    
    
    
def test_estimate_computation_speed():    
    """ test estimate_computation_speed method """
    def f(x):
        return 2 * x
    
    def g(x):
        return np.sin(x) * np.cos(x)**2
    
    assert (misc.estimate_computation_speed(f, 1) >
            misc.estimate_computation_speed(g, 1))
    
    
    
def test_classproperty():
    """ test classproperty decorator """
    class Test():
        _value = 2
        
        @misc.classproperty
        def value(cls):  # @NoSelf
            return cls._value
        
    assert Test.value == 2
    
    
    
def test_progress_bars():
    """ test progress bars """
    for pb_cls in [misc.MockProgress, misc.get_progress_bar_class()]:
        tot = 0
        for i in pb_cls(range(4)):
            tot += i
        assert tot == 6
    
    
    
@misc.skipUnlessModule('h5py')
def test_hdf_write_attributes(tmp_path):
    """ test hdf_write_attributes function """
    import h5py
    path = tmp_path / "test_hdf_write_attributes.hdf5"

    # test normal case    
    data = {'a': 3, 'b': 'asd'}
    with h5py.File(path, 'w') as hdf_file:
        misc.hdf_write_attributes(hdf_file, data)
        data2 = {k: json.loads(v) for k, v in hdf_file.attrs.items()}
            
    assert data == data2
    assert data is not data2

    # test silencing of problematic items
    with h5py.File(path, 'w') as hdf_file:
        misc.hdf_write_attributes(hdf_file, {'a': 1, 'b': object()})
        data2 = {k: json.loads(v) for k, v in hdf_file.attrs.items()}
    assert data2 == {'a': 1}
            
    # test raising problematic items
    with h5py.File(path, 'w') as hdf_file:
        with pytest.raises(TypeError):
            misc.hdf_write_attributes(hdf_file, {'a': object()},
                                      raise_serialization_error=True)
                


'''
This module provides infrastructure for managing classes with parameters. One
aim is to allow easy management of inheritance of parameters.

.. autosummary::
   :nosignatures:

   Parameter
   DeprecatedParameter
   HideParameter
   Parameterized
   get_all_parameters

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''


import logging
from collections import OrderedDict
from typing import Sequence, Dict, Any, Union

from pde.tools.misc import import_class, hybridmethod, in_jupyter_notebook



class Parameter():
    """ class representing a single parameter """
    
    def __init__(self, name: str,
                 default_value=None,
                 cls=object,
                 description: str = '',
                 hidden: bool = False):
        """ initialize a parameter 
        
        Args:
            name (str):
                The name of the parameter
            default_value:
                The default value
            cls:
                The type of the parameter, which is used for conversion
            description (str):
                A string describing the impact of this parameter. This
                description appears in the parameter help
            hidden (bool):
                Whether the parameter is hidden in the description summary
        """
        self.name = name
        self.default_value = default_value
        self.cls = cls
        self.description = description
        self.hidden = hidden
        if cls is not object and cls(default_value) != default_value:
            logging.warning('Default value `%s` does not seem to be of type '
                            '`%s`', name, cls.__name__)
    
        
    def __repr__(self):
        return (f'{self.__class__.__name__}(name="{self.name}", default_value='
                f'"{self.default_value}", cls="{self.cls.__name__}", '
                f'description="{self.description}", hidden={self.hidden})')
    __str__ = __repr__
    
    
    def __getstate__(self):
        # replace the object class by its class path 
        return {'name': str(self.name),
                'default_value': self.convert(),
                'cls': object.__module__ + '.' + self.cls.__name__,
                'description': self.description,
                'hidden': self.hidden}
    
        
    def __setstate__(self, state):
        # restore the object from the class path
        state['cls'] = import_class(state['cls'])
        # restore the state
        self.__dict__.update(state)
        
        
    def convert(self, value=None):
        """ converts a `value` into the correct type for this parameter. If
        `value` is not given, the default value is converted.
        
        Note that this does not make a copy of the values, which could lead to
        unexpected effects where the default value is changed by an instance.
        
        Args:
            value: The value to convert
        
        Returns:
            The converted value, which is of type `self.cls`
        """ 
        if value is None:
            value = self.default_value
            
        if self.cls is object:
            return value
        else:
            try:
                return self.cls(value)
            except ValueError:
                raise ValueError(f"Could not convert {value!r} to "
                                 f"{self.cls.__name__} for parameter "
                                 f"'{self.name}'")



class DeprecatedParameter(Parameter):
    """ a parameter that can still be used normally but is deprecated """
    pass



class HideParameter():
    """ a helper class that allows hiding parameters of the parent classes """
    
    
    def __init__(self, name: str):
        """ 
        Args:
            name (str):
                The name of the parameter
        """
        self.name = name



ParameterListType = Sequence[Union[Parameter, HideParameter]]



class Parameterized():
    """ a mixin that manages the parameters of a class """

    parameters_default: ParameterListType = []
    _subclasses: Dict[str, 'Parameterized'] = {}


    def __init__(self, parameters: Dict[str, Any] = None):
        """ initialize the parameters of the object
        
        Args:
            parameters (dict):
                A dictionary of parameters to change the defaults. The allowed
                parameters can be obtained from
                :meth:`~Parameterized.get_parameters` or displayed by calling
                :meth:`~Parameterized.show_parameters`.
        """
        # set logger if this has not happened, yet
        if not hasattr(self, '_logger'):
            self._logger = logging.getLogger(self.__class__.__name__)
            
        # set parameters if they have not been initialized, yet
        if not hasattr(self, 'parameters'):
            self.parameters = self._parse_parameters(parameters)


    def __init_subclass__(cls, **kwargs):  # @NoSelf
        """ register all subclasses to reconstruct them later """
        # normalize the parameters_default attribute
        if (hasattr(cls, 'parameters_default') and
                isinstance(cls.parameters_default, dict)):
            # default parameters are given as a dictionary
            cls.parameters_default = \
                [Parameter(*args) for args in cls.parameters_default.items()]
        
        # register the subclasses
        super().__init_subclass__(**kwargs)
        cls._subclasses[cls.__name__] = cls
            

    @classmethod
    def get_parameters(cls, include_hidden: bool = False,
                       include_deprecated: bool = False,
                       sort: bool = True) -> Dict[str, Parameter]:
        """ return a dictionary of parameters that the class supports
        
        Args:
            include_hidden (bool): Include hidden parameters
            include_deprecated (bool): Include deprecated parameters
            sort (bool): Return ordered dictionary with sorted keys
            
        Returns:
            dict: a dictionary of instance of :class:`Parameter` with their
            names as keys.
        """
        # collect the parameters from the class hierarchy
        parameters: Dict[str, Parameter] = {}
        for cls in reversed(cls.__mro__):
            if hasattr(cls, 'parameters_default'):
                for p in cls.parameters_default:
                    if isinstance(p, HideParameter):
                        if include_hidden:
                            parameters[p.name].hidden = True
                        else:
                            del parameters[p.name]
                            
                    else:
                        parameters[p.name] = p
                        
        # filter parameters based on hidden and deprecated flags
        def show(p):
            """ helper function to decide whether parameter will be shown """
            # show based on hidden flag?
            show1 = include_hidden or not p.hidden
            # show based on deprecated flag?
            show2 = include_deprecated or not isinstance(p, DeprecatedParameter)
            return show1 and show2
        
        # filter parameters based on `show`
        result = {name: parameter
                  for name, parameter in parameters.items()
                  if show(parameter)}
            
        if sort:
            result = OrderedDict(sorted(result.items()))
        return result
            
        
    @classmethod    
    def _parse_parameters(cls, parameters: Dict[str, Any] = None,
                          check_validity: bool = True,
                          allow_hidden: bool = True,
                          include_deprecated: bool = False) -> Dict[str, Any]:
        """ parse parameters

        Args:
            parameters (dict):
                A dictionary of parameters that will be parsed. 
            check_validity (bool):
                Determines whether a `ValueError` is raised if there are keys in
                parameters that are not in the defaults. If `False`, additional
                items are simply stored in `self.parameters`
            allow_hidden (bool):
                Allow setting hidden parameters
            include_deprecated (bool):
                Include deprecated parameters
        """
        if parameters is None:
            parameters = {}
        else:
            parameters = parameters.copy()  # do not modify the original
        
        # obtain all possible parameters
        param_objs = cls.get_parameters(include_hidden=allow_hidden,
                                        include_deprecated=include_deprecated)
        
        # initialize parameters with default ones from all parent classes
        result: Dict[str, Any] = {}
        for name, param_obj in param_objs.items():
            if not allow_hidden and param_obj.hidden:
                continue  # skip hidden parameters
            # take value from parameters or set default value
            result[name] = param_obj.convert(parameters.pop(name, None))
                
        # update parameters with the supplied ones
        if check_validity and parameters:
            raise ValueError(f'Parameters `{sorted(parameters.keys())}` were '
                             'provided in instance specific parameters but are '
                             f'not defined for the class `{cls.__name__}`.')
        else:
            result.update(parameters)  # add remaining parameters

        return result
            

    def get_parameter_default(self, name):
        """ return the default value for the parameter with `name` 
        
        Args:
            name (str): The parameter name
        """
        for cls in self.__class__.__mro__:
            if hasattr(cls, 'parameters_default'):
                for p in cls.parameters_default:
                    if isinstance(p, Parameter) and p.name == name:
                        return p.default_value

        raise KeyError(f'Parameter `{name}` is not defined')
        
        
    @classmethod
    def _show_parameters(cls, description: bool = None,
                         sort: bool = False,
                         show_hidden: bool = False,
                         show_deprecated: bool = False,
                         parameter_values: Dict[str, Any] = None):
        """ private method showing all parameters in human readable format
        
        Args:
            description (bool):
                Flag determining whether the parameter description is shown. The
                default is to show the description only when we are in a jupyter
                notebook environment.
            sort (bool):
                Flag determining whether the parameters are sorted
            show_hidden (bool):
                Flag determining whether hidden parameters are shown
            show_deprecated (bool):
                Flag determining whether deprecated parameters are shown
            parameter_values (dict):
                A dictionary with values to show. Parameters not in this
                dictionary are shown with their default value.
        
        All flags default to `False`.
        """
        # determine whether we are in a jupyter notebook and can return HTML
        in_notebook = in_jupyter_notebook()
        try:
            from IPython.display import HTML, display
        except ImportError:
            in_notebook = False
        
        if description is None:
            description = in_notebook  # show only in notebook by default
        
        # set the templates for displaying the data 
        if in_notebook:
            # templates for HTML output
            template = '<dt>{name} = {value!r}</dt>'
            if description:
                template += '<dd>{description}</dd>'
            html = ''

        else:
            # template for normal output
            template = '{name}: {type} = {value!r}'
            template_object = '{name} = {value!r}'
            if description:
                template += ' ({description})'
                template_object += ' ({description})'
            
        # iterate over all parameters
        params = cls.get_parameters(include_hidden=show_hidden,
                                    include_deprecated=show_deprecated,
                                    sort=sort)
        for param in params.values():
            # initialize the data to show
            data = {'name': param.name,
                    'type': param.cls.__name__,
                    'description': param.description}
            
            # determine the value to show
            if parameter_values is None:
                data['value'] = param.default_value
            else:
                data['value'] = parameter_values[param.name]
            
            if in_notebook:
                # collect the data for later printing as HTML
                html += template.format(**data)
            else:
                # print the data to stdout
                if param.cls is object:
                    print((template_object.format(**data)))
                else:
                    print((template.format(**data)))
            
        if in_notebook:
            # output html with minimal styling
            css = 'dl.py-pde_parameters dd {padding-left:2em}'
            display(HTML(f'<style type="text/css">{css}</style>'
                         f'<dl class="py-pde_parameters">{html}</dl>'))
            

    @hybridmethod
    def show_parameters(cls, description: bool = None,  # @NoSelf
                        sort: bool = False,
                        show_hidden: bool = False,
                        show_deprecated: bool = False):
        """ show all parameters in human readable format
        
        Args:
            description (bool):
                Flag determining whether the parameter description is shown. The
                default is to show the description only when we are in a jupyter
                notebook environment.
            sort (bool):
                Flag determining whether the parameters are sorted
            show_hidden (bool):
                Flag determining whether hidden parameters are shown
            show_deprecated (bool):
                Flag determining whether deprecated parameters are shown
        
        All flags default to `False`.
        """
        cls._show_parameters(description, sort, show_hidden, show_deprecated)    


    @show_parameters.instancemethod  # type: ignore
    def show_parameters(self, description: bool = None,  # @NoSelf
                        sort: bool = False,
                        show_hidden: bool = False,
                        show_deprecated: bool = False,
                        default_value: bool = False):
        """ show all parameters in human readable format
        
        Args:
            description (bool):
                Flag determining whether the parameter description is shown. The
                default is to show the description only when we are in a jupyter
                notebook environment.
            sort (bool):
                Flag determining whether the parameters are sorted
            show_hidden (bool):
                Flag determining whether hidden parameters are shown
            show_deprecated (bool):
                Flag determining whether deprecated parameters are shown
            default_value (bool):
                Flag determining whether the default values or the current
                values are shown
        
        All flags default to `False`.
        """
        self._show_parameters(description, sort, show_hidden, show_deprecated,
                              None if default_value else self.parameters)
        
        

def get_all_parameters(data: str = 'name') -> Dict[str, Any]:
    """ get a dictionary with all parameters of all registered classes
    
    Args:
        data (str):
            Determines what data is returned. Possible values are 'name',
            'value', or 'description', to return the respective information
            about the parameters.
    """
    result = {}
    for cls_name, cls in Parameterized._subclasses.items():
        if data == 'name':
            parameters = set(cls.get_parameters().keys())
        elif data == 'value':
            parameters = {k: v.default_value  # type: ignore
                          for k, v in cls.get_parameters().items()}
        elif data == 'description':
            parameters = {k: v.description  # type: ignore
                          for k, v in cls.get_parameters().items()}
        else:
            raise ValueError(f'Cannot interpret data `{data}`')
        
        result[cls_name] = parameters
    return result
        
        

def sphinx_display_parameters(app, what, name, obj, options, lines):
    """ helper function to display parameters in sphinx documentation
    
    Example: 
        This function should be connected to the 'autodoc-process-docstring'
        event like so:
    
            app.connect('autodoc-process-docstring', sphinx_display_parameters)
    """
    if what == 'class' and issubclass(obj, Parameterized):
        if any(':param parameters:' in line for line in lines):
            # parse parameters
            parameters = obj.get_parameters(sort=False)
            if parameters:
                lines.append('.. admonition::')
                lines.append(f'   Parameters of {obj.__name__}:')
                lines.append('   ')
                for p in parameters.values():
                    lines.append(f'   {p.name}')
                    text = p.description.splitlines()
                    text.append(f'(Default value: :code:`{p.default_value!r}`)')
                    text = ['     ' + t for t in text]
                    lines.extend(text)
                    lines.append('')
                lines.append('')



"""
Simple sphinx plug-in that simplifies  type information in function signatures
"""

import collections
import re


# simple (literal) replacement rules
REPLACEMENTS = collections.OrderedDict([
    # Remove some unnecessary information in favor of a more compact style 
    ('Dict[KT, VT]', 'dict'),
    ('Dict[str, Any]', 'dict'),
    ('Optional[str]', 'str'),
    ('Optional[float]', 'float'),
    ('Optional[int]', 'int'),
    ('Optional[dict]', 'dict'),
    ('Optional[Dict[str, Any]]', 'dict'),
    # Complex types that can be represented by some descriptive name
    ('Union[dict, str, BCBase]', 'BoundaryData'),
    ('Union[Dict[str, BoundaryData], '
     'BoundaryData, Tuple[BoundaryData, BoundaryData]]', 'BoundaryPairData'),
    ('Union[Dict[str, BoundaryData], dict, str, BCBase, '
     'Tuple[BoundaryData, BoundaryData]]', 'BoundaryPairData'),
    ('Union[BoundaryPairData, Sequence[BoundaryPairData]]', 'BoundariesData'),
    ('Union[Dict[str, BoundaryData], dict, str, BCBase, '
     'Tuple[BoundaryData, BoundaryData], Sequence[BoundaryPairData]]',
        'BoundariesData'),
    ('Union[dict, str, BCBase, Tuple[Union[dict, str, BCBase], '
     'Union[dict, str, BCBase]], Sequence[Union[dict, str, BCBase, '
     'Tuple[Union[dict, str, BCBase], Union[dict, str, BCBase]]]]]',
        'BoundaryConditionData'),
    ('Union[Dict[str, Union[Dict, str, BCBase]], Dict, str, BCBase, '
     'Tuple[Union[Dict, str, BCBase], Union[Dict, str, BCBase]], '
     'Sequence[Union[Dict[str, Union[Dict, str, BCBase]], Dict, str, BCBase, '
     'Tuple[Union[Dict, str, BCBase], Union[Dict, str, BCBase]]]]]',
        'BoundaryConditionData'),
    ('Union[dict, str, BCBase, Tuple[Union[dict, str, BCBase], '
     'Union[dict, str, BCBase]], Sequence[Union[dict, str, BCBase, '
     'Tuple[Union[dict, str, BCBase], Union[dict, str, BCBase]]]], '
     'Sequence[BoundaryConditionData]]',
        'BoundaryConditionData'),
    ('Union[Dict[str, BoundaryData], dict, str, BCBase, '
     'Tuple[BoundaryData, BoundaryData], '
     'Sequence[BoundaryPairData], Sequence[BoundariesData]]',
        'BoundariesDataList'),
    ('Union[Dict[str, Union[Dict, str, BCBase]], Dict, str, BCBase, '
     'Tuple[Union[Dict, str, BCBase], Union[Dict, str, BCBase]], '
     'Sequence[Union[Dict[str, Union[Dict, str, BCBase]], Dict, str, BCBase, '
     'Tuple[Union[Dict, str, BCBase], Union[Dict, str, BCBase]]]], '
     'Sequence[BoundaryConditionData]]',
        'BoundariesDataList'),
    ('Union[List[Union[TrackerBase, str]], TrackerBase, str, None]',
        'TrackerData'),
])


# replacement rules based on regular expressions
REPLACEMENTS_REGEX = {
    # remove full package path and only leave the module/class identifier
    'pde\.(\w+\.)*': '',
}



def process_signature(app, what: str, name: str, obj, options, signature,
                      return_annotation):
    """ Process signature by applying replacement rules """
    if signature is not None:
        for key, value in REPLACEMENTS.items():
            signature = signature.replace(key, value)
        for key, value in REPLACEMENTS_REGEX.items():
            signature = re.sub(key, value, signature)
    return signature, return_annotation



def setup(app):
    """ set up hooks for this sphinx plugin """ 
    app.connect('autodoc-process-signature', process_signature)
    


#!/usr/bin/env python3

import sys
from pathlib import Path

PACKAGE_PATH = Path(__file__).resolve().parents[1]
sys.path.append(str(PACKAGE_PATH))

import numpy as np

from pde import UnitGrid, ScalarField
from pde.grids.boundaries.axes import Boundaries
from pde.tools.misc import estimate_computation_speed



def main():
    """ main routine testing the performance """
    print('Reports calls-per-second (larger is better)\n')
    
    # Cartesian grid with different shapes and boundary conditions
    for size in [32, 512]:
        grid = UnitGrid((size, size), periodic=False)
        print(grid)
        
        field = ScalarField.random_normal(grid)
        bc_value = np.ones(size)
        result = field.laplace(bc={'value': 1}).data
        
        for bc in ['scalar', 'array', 'linked']:
            if bc == 'scalar':
                bcs = {'value': 1}
            elif bc == 'array':
                bcs = {'value': bc_value}
            elif bc == 'linked':
                bcs = Boundaries.from_data(grid, {'value': bc_value}, rank=0)
                for ax, upper in grid._iter_boundaries():
                    bcs[ax][upper].link_value(bc_value)
            #result = field.laplace(bc=bcs).data
            laplace = grid.get_operator("laplace", bc=bcs)
            # call once to pre-compile and test result 
            np.testing.assert_allclose(laplace(field.data), result)
            speed = estimate_computation_speed(laplace, field.data)
            print(f'{bc:>6s}:{int(speed):>9d}')
            
        print()
 


if __name__ == '__main__':
    main()



"""
Using simulation trackers
=========================

This example illustrates how trackers can be used to analyze simulations.
"""

from pde import (DiffusionPDE, UnitGrid, ScalarField, MemoryStorage,
                 PlotTracker, PrintTracker, RealtimeIntervals)

grid = UnitGrid([16, 16])  # generate grid
state = ScalarField.random_uniform(grid, 0.2, 0.3)  # generate initial condition

storage = MemoryStorage()

trackers = [
    'progress',                    # show progress bar during simulation
    'steady_state',                # abort when steady state is reached
    storage.tracker(interval=1),   # store data every simulation time unit
    PlotTracker(show=True),        # show images during simulation
    # print some output every 5 real seconds:
    PrintTracker(interval=RealtimeIntervals(duration=5))
]

eq = DiffusionPDE()  # define the PDE
eq.solve(state, 10, dt=0.1, tracker=trackers)


for field in storage:
    print(field.integral)



r"""

This package contains classes for handling the boundary conditions of fields.

.. _documentation-boundaries:

Boundary conditions
^^^^^^^^^^^^^^^^^^^

Since the :mod:`pde` package only supports orthogonal grids, boundary conditions
need to be applied at the end of each axis. Consequently, methods expecting 
boundary conditions typically receive a list of conditions for each axes:

.. code-block:: python

    field = ScalarField(UnitGrid([16, 16], periodic=[True, False]))
    field.laplace(bc=[bc_x, bc_y])
    
If an axis is periodic (like the first one in the example above), the only valid
boundary condition is 'periodic'. For non-periodic axes (e.g., the second axis),
different boundary conditions can be specified for the lower and upper end of
the axis, which is done using a tuple of two conditions. Typical choices 
for individual conditions are Dirichlet conditions that enforce a value
NUM (specified by `{'value': NUM}`) and Neumann conditions that enforce 
the value DERIV for the derivative in the normal direction (specified by
`{'derivative': DERIV}`). The specific choices for the example above could be

.. code-block:: python
    
    bc_x = 'periodic'
    bc_y = ({'value': 2}, {'derivative': -1})

which enforces a value of `2` at the lower side of the y-axis and a derivative
(in outward normal direction) of `-1` on the upper side. Instead of plain
numbers, enforcing the same condition along the whole boundary, expressions are
also supported. These are given as mathematical equations that can be parsed by
sympy. These expressions can depend on all coordinates of the grid, except the
coordinate associated with the particular boundary. An alternative boundary
condition to the example above could thus read

.. code-block:: python
    
    bc_y = ({'value': 'y**2'}, {'derivative': '-sin(x)'})


Warning:
    To interpret arbitrary expressions, the package uses :func:`exec`. It
    should therefore not be used in a context where malicious input could occur.
    

The package also supports mixed boundary conditions (depending on both the value
and the derivative of the field) and imposing a second derivative. An example is

.. code-block:: python
    
    bc_y = ({'type': 'mixed', 'value': 2, 'const': 7},
            {'curvature': 2})

which enforces the condition :math:`\partial_n c + 2 c = 7` and
:math:`\partial^2_n c = 2` onto the field :math:`c` on the lower and upper side
of the axis, respectively.

Beside the full specification of the boundary condtions, various short-hand
notations are supported. If both sides of an axis have the same boundary
condition, only one needs to be specified (instead of the tuple). For instance,
:code:`bc_y = {'value': 2}` imposes a value of `2` on both sides of the y-axis.
Similarly, if all axes have the same boundary conditions, only one axis needs to
be specified (instead of the list). For instance, the following example

.. code-block:: python

    field = ScalarField(UnitGrid([16, 16], periodic=False))
    field.laplace(bc={'value': 2})

imposes a value of `2` on all sides of the grid. Finally, the special value
'natural' imposes periodic boundary conditions for periodic axis and a vanishing
derivative otherwise. For example, 

.. code-block:: python

    field = ScalarField(UnitGrid([16, 16], periodic=[True, False]))
    field.laplace(bc='natural')
    
enforces periodic boundary conditions on the first axis, while the second one
has standard Neumann conditions.
    
Note:
    Derivatives are given relative to the outward normal vector, such that
    positive derivatives correspond to a function that increases across the
    boundary, which corresponds to an inwards flux. Conversely, negative
    derivatives are associated with effluxes. 


Boundaries overview
^^^^^^^^^^^^^^^^^^^

The :mod:`boundaries` package defines the following classes:

**Local boundary conditions:**

* :class:`~pde.grids.boundaries.local.DirichletBC`:
  Imposing the value of a field at the boundary
* :class:`~pde.grids.boundaries.local.NeumannBC`:
  Imposing the derivative of a field in the outward normal direction at the
  boundary
* :class:`~pde.grids.boundaries.local.MixedBC`:
  Imposing the derivative of a field in the outward normal direction
  proportional to its value at the boundary  
* :class:`~pde.grids.boundaries.local.CurvatureBC`:
  Imposing the second derivative (curvature) of a field at the boundary
* :class:`~pde.grids.boundaries.local.ExtrapolateBC`:
  Extrapolate boundary points linearly from the two points closest to the
  boundary

**Boundaries for an axis:**

* :class:`~pde.grids.boundaries.axis.BoundaryPair`:
  Uses the local boundary conditions to specify the two boundaries along an
  axis.
* :class:`~pde.grids.boundaries.axis.BoundaryPeriodic`:
  Indicates that an axis has periodic boundary conditions

**Boundaries for all axes of a grid:**

* :class:`~pde.grids.boundaries.axes.Boundaries`:
  Collection of boundaries to describe conditions for all axes


**Inheritance structure of the classes:**

.. inheritance-diagram:: pde.grids.boundaries.axes pde.grids.boundaries.axis
        pde.grids.boundaries.local
   :parts: 2

The details of the classes are explained below:

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from .. import DomainError, PeriodicityError
from .axes import Boundaries



r"""
This module implements differential operators on polar grids 

.. autosummary::
   :nosignatures:

   make_laplace
   make_gradient
   make_divergence
   make_vector_gradient
   make_tensor_divergence
   
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from typing import Callable

from scipy import sparse

from .common import make_general_poisson_solver
from .. import PolarGrid
from ..boundaries import Boundaries
from ...tools.numba import jit_allocate_out
from ...tools.docstrings import fill_in_docstring



@fill_in_docstring
def make_laplace(bcs: Boundaries) -> Callable:
    """ make a discretized laplace operator for a polar grid
    
    {DESCR_POLAR_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, PolarGrid)
    bcs.check_value_rank(0)
    
    # calculate preliminary quantities
    dim_r = bcs.grid.shape[0]
    dr = bcs.grid.discretization[0]
    rs = bcs.grid.axes_coords[0]
    r_min, _ = bcs.grid.axes_bounds[0]
    dr_2 = 1 / dr**2

    # prepare boundary values
    value_lower_bc = bcs[0].low.make_virtual_point_evaluator()
    value_upper_bc = bcs[0].high.make_virtual_point_evaluator()    
    
    @jit_allocate_out(out_shape=(dim_r,))
    def laplace(arr, out=None):
        """ apply laplace operator to array `arr` """
        i = 0
        if r_min == 0:
            out[i] = 2 * (arr[i + 1] - arr[i]) * dr_2
        else:
            arr_r_l = value_lower_bc(arr, (i,))
            out[i] = ((arr[i + 1] - 2 * arr[i] + arr_r_l) * dr_2 +
                      (arr[i + 1] - arr_r_l) / (2 * rs[i] * dr))
        
        for i in range(1, dim_r - 1):  # iterate inner radial points
            out[i] = ((arr[i + 1] - 2 * arr[i] + arr[i - 1]) * dr_2 +
                      (arr[i + 1] - arr[i - 1]) / (2 * rs[i] * dr))
            
        # express boundary condition at outer side
        i = dim_r - 1
        arr_r_h = value_upper_bc(arr, (i,))
        out[i] = ((arr_r_h - 2 * arr[i] + arr[i - 1]) * dr_2 +
                  (arr_r_h - arr[i - 1]) / (2 * rs[i] * dr))
        return out      
          
    return laplace  # type: ignore



@fill_in_docstring
def make_gradient(bcs: Boundaries) -> Callable:
    """ make a discretized gradient operator for a polar grid
    
    {DESCR_POLAR_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, PolarGrid)
    bcs.check_value_rank(0)

    # calculate preliminary quantities
    dim_r = bcs.grid.shape[0]
    r_min, _ = bcs.grid.axes_bounds[0]
    dr = bcs.grid.discretization[0]
    scale_r = 1 / (2 * dr)
    
    # prepare boundary values
    boundary = bcs[0]
    value_lower_bc = boundary.low.make_virtual_point_evaluator()
    value_upper_bc = boundary.high.make_virtual_point_evaluator()
    
    @jit_allocate_out(out_shape=(2, dim_r))
    def gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        # no-flux at the origin 
        i = 0
        if r_min == 0:
            out[0, i] = (arr[1] - arr[0]) * scale_r
        else:
            arr_r_l = value_lower_bc(arr, (i,))
            out[0, i] = (arr[1] - arr_r_l) * scale_r            
        out[1, i] = 0  # no angular dependence by definition
        
        for i in range(1, dim_r - 1):  # iterate inner radial points
            out[0, i] = (arr[i + 1] - arr[i - 1]) * scale_r
            out[1, i] = 0  # no angular dependence by definition

        i = dim_r - 1
        arr_r_h = value_upper_bc(arr, (i,))
        out[0, i] = (arr_r_h - arr[i - 1]) * scale_r
        out[1, i] = 0  # no angular dependence by definition
        
        return out
        
    return gradient  # type: ignore



@fill_in_docstring
def make_divergence(bcs: Boundaries) -> Callable:
    """ make a discretized divergence operator for a polar grid
    
    {DESCR_POLAR_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, PolarGrid)
    bcs.check_value_rank(0)

    # calculate preliminary quantities
    dim_r = bcs.grid.shape[0]
    dr = bcs.grid.discretization[0]
    rs = bcs.grid.axes_coords[0]
    r_min, _ = bcs.grid.axes_bounds[0]
    scale_r = 1 / (2 * dr)
    
    # prepare boundary values
    boundary = bcs[0]
    value_lower_bc = boundary.low.make_virtual_point_evaluator()
    value_upper_bc = boundary.high.make_virtual_point_evaluator()

    if r_min == 0:
        @jit_allocate_out(out_shape=(dim_r,))
        def divergence(arr, out=None):
            """ apply divergence operator to array `arr` """            
            # inner radial boundary condition
            i = 0
            out[i] = (arr[0, 1] + 3 * arr[0, 0]) * scale_r
            
            for i in range(1, dim_r - 1):  # iterate radial points
                out[i] = ((arr[0, i + 1] - arr[0, i - 1]) * scale_r + 
                          (arr[0, i] / ((i + 0.5) * dr)))
                
            # outer radial boundary condition
            i = dim_r - 1
            arr_r_h = value_upper_bc(arr[0], (i,))
            out[i] = ((arr_r_h - arr[0, i - 1]) * scale_r + 
                      (arr[0, i] / ((i + 0.5) * dr)))
                
            return out
        
    else:  # r_min > 0
        @jit_allocate_out(out_shape=(dim_r,))
        def divergence(arr, out=None):
            """ apply divergence operator to array `arr` """            
            # inner radial boundary condition
            i = 0
            arr_r_l = value_lower_bc(arr[0], (i,))
            out[i] = (arr[0, i + 1] - arr_r_l) * scale_r + arr[0, i] / rs[i]
            
            for i in range(1, dim_r - 1):  # iterate radial points
                out[i] = ((arr[0, i + 1] - arr[0, i - 1]) * scale_r + 
                          arr[0, i] / rs[i])
                
            # outer radial boundary condition
            i = dim_r - 1
            arr_r_h = value_upper_bc(arr[0], (i,))
            out[i] = (arr_r_h - arr[0, i - 1]) * scale_r + arr[0, i] / rs[i]
                
            return out
        
    return divergence  # type: ignore

    
    
@fill_in_docstring
def make_vector_gradient(bcs: Boundaries) -> Callable:
    """ make a discretized vector gradient operator for a polar grid
    
    {DESCR_POLAR_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, PolarGrid)
    bcs.check_value_rank(1)

    gradient_r = make_gradient(bcs.extract_component(0))
    gradient_phi = make_gradient(bcs.extract_component(1))
        
    @jit_allocate_out(out_shape=(2, 2) + bcs.grid.shape)
    def vector_gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        gradient_r(arr[0], out=out[:, 0])
        gradient_phi(arr[1], out=out[:, 1])
        return out    
        
    return vector_gradient  # type: ignore



@fill_in_docstring
def make_tensor_divergence(bcs: Boundaries) -> Callable:
    """ make a discretized tensor divergence operator for a polar grid
    
    {DESCR_POLAR_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, PolarGrid)
    bcs.check_value_rank(1)

    divergence_r = make_divergence(bcs.extract_component(0))
    divergence_phi = make_divergence(bcs.extract_component(1))
        
    @jit_allocate_out(out_shape=(2,) + bcs.grid.shape)
    def tensor_divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        divergence_r(arr[0], out=out[0])
        divergence_phi(arr[1], out=out[1])
        return out
        
    return tensor_divergence  # type: ignore



@fill_in_docstring
def _get_laplace_matrix(bcs):
    """ get sparse matrix for laplace operator on a polar grid
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        tuple: A sparse matrix and a sparse vector that can be used to evaluate
        the discretized laplacian
    """    
    assert isinstance(bcs.grid, PolarGrid)
    bcs.check_value_rank(0)
    
    # calculate preliminary quantities
    dim_r = bcs.grid.shape[0]
    dr = bcs.grid.discretization[0]
    rs = bcs.grid.axes_coords[0]
    r_min, _ = bcs.grid.axes_bounds[0]
    scale = 1 / dr**2

    matrix = sparse.dok_matrix((dim_r, dim_r))
    vector = sparse.dok_matrix((dim_r, 1))
    
    for i in range(dim_r):
        matrix[i, i] += -2 * scale
        scale_i = 1 / (2 * rs[i] * dr)
        
        if i == 0:
            if r_min == 0:
                matrix[i, i + 1] = 2 * scale
                continue  # the special case of the inner boundary is handled
            else:
                const, entries = bcs[0].get_data((-1,))
                factor = scale - scale_i
                vector[i] += const * factor
                for k, v in entries.items():
                    matrix[i, k] += v * factor

        else:
            matrix[i, i - 1] = scale - scale_i
            
        if i == dim_r - 1:
            const, entries = bcs[0].get_data((dim_r,))
            factor = scale + scale_i
            vector[i] += const * factor
            for k, v in entries.items():
                matrix[i, k] += v * factor
                
        else:
            matrix[i, i + 1] = scale + scale_i
            
    return matrix, vector



@fill_in_docstring
def make_poisson_solver(bcs: Boundaries, method: str = 'auto') -> Callable:
    """ make a operator that solves Poisson's equation
    
    {DESCR_POLAR_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    matrix, vector = _get_laplace_matrix(bcs)
    return make_general_poisson_solver(matrix, vector, method)

       
# register all operators with the grid class
PolarGrid.register_operator('laplace', make_laplace,
                            rank_in=0, rank_out=0)
PolarGrid.register_operator('gradient', make_gradient,
                            rank_in=0, rank_out=1)
PolarGrid.register_operator('divergence', make_divergence,
                            rank_in=1, rank_out=0)
PolarGrid.register_operator('vector_gradient', make_vector_gradient,
                            rank_in=1, rank_out=2)
PolarGrid.register_operator('tensor_divergence', make_tensor_divergence,
                            rank_in=2, rank_out=1)
PolarGrid.register_operator('poisson_solver', make_poisson_solver,
                            rank_in=0, rank_out=0)



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import itertools

import numpy as np
import pytest

from .test_generic import iter_grids
from .. import ScalarField, VectorField, Tensor2Field, FieldCollection
from ..base import FieldBase
from ...grids import UnitGrid



def test_shapes_nfields():
    """ test single component field """
    for num, grid in itertools.product([1, 3], iter_grids()):
        fields = [ScalarField.random_uniform(grid)
                  for _ in range(num)]
        field = FieldCollection(fields)
        data_shape = (num, ) + grid.shape
        np.testing.assert_equal(field.data.shape, data_shape)
        for pf_single in field:
            np.testing.assert_equal(pf_single.data.shape, grid.shape)
         
        field_c = field.copy()
        np.testing.assert_allclose(field.data, field_c.data)
        assert field.grid == field_c.grid



def test_collections():
    """ test field collections """
    grid = UnitGrid([3, 4])
    sf = ScalarField.random_uniform(grid, label='sf')
    vf = VectorField.random_uniform(grid, label='vf')
    tf = Tensor2Field.random_uniform(grid, label='tf')
    fields = FieldCollection([sf, vf, tf])
    assert fields.data.shape == (7, 3, 4)
    assert isinstance(str(fields), str)
    
    fields.data[:] = 0
    np.testing.assert_allclose(sf.data, 0)
    np.testing.assert_allclose(vf.data, 0)
    np.testing.assert_allclose(tf.data, 0)
    
    assert fields[0] is fields['sf']
    assert fields[1] is fields['vf']
    assert fields[2] is fields['tf']
    with pytest.raises(KeyError):
        fields['42']

    sf.data = 1
    vf.data = 1
    tf.data = 1
    np.testing.assert_allclose(fields.data, 1)
    
    assert sf.data.shape == (3, 4)
    assert vf.data.shape == (2, 3, 4)
    assert tf.data.shape == (2, 2, 3, 4)
    
    c2 = FieldBase.from_state(fields.attributes, data=fields.data)
    assert c2 == fields
    assert c2.grid is grid
    
    attrs = FieldCollection.unserialize_attributes(fields.attributes_serialized)
    c2 = FieldCollection.from_state(attrs, data=fields.data)
    assert c2 == fields
    assert c2.grid is not grid
    
    fields['sf'] = 2.
    np.testing.assert_allclose(sf.data, 2)
    with pytest.raises(KeyError):
        fields['42'] = 0
    
    
def test_collections_copy():
    """ test copying data of collections """
    grid = UnitGrid([2, 2])
    sf = ScalarField(grid, 0)
    vf = VectorField(grid, 1)
    fc = FieldCollection([sf, vf])
    
    data = np.r_[np.zeros(4), np.ones(8)]
    np.testing.assert_allclose(fc.data.flat, data)
    
    fc2 = fc.copy()
    assert fc.data is not fc2.data
    assert fc[0].data is not fc2[0].data
    assert fc[1].data is not fc2[1].data
    
    sf.data = 1
    np.testing.assert_allclose(fc.data.flat, np.ones(12))
    np.testing.assert_allclose(fc2.data.flat, data)


def test_collections_operators():
    """ test field collections """
    grid = UnitGrid([3, 4])
    sf = ScalarField(grid, 1)
    vf = VectorField(grid, 1)
    fields = FieldCollection([sf, vf])
    
    fields += fields
    np.testing.assert_allclose(fields.data, 2)
    np.testing.assert_allclose(sf.data, 2)
    np.testing.assert_allclose(vf.data, 2)
    
    fields = fields - 1
    np.testing.assert_allclose(fields.data, 1)
    
    fields = fields + fields
    np.testing.assert_allclose(fields.data, 2)
     
    fields *= 2
    np.testing.assert_allclose(fields.data, 4)
            
            
            
def test_smoothing_collection():
    """ test smoothing of a FieldCollection """
    grid = UnitGrid([3, 4], periodic=[True, False])
    sf = ScalarField.random_uniform(grid)
    vf = VectorField.random_uniform(grid)
    tf = Tensor2Field.random_uniform(grid)
    fields = FieldCollection([sf, vf, tf])
    sgm = 0.5 + np.random.random()

    out = fields.smooth(sigma=sgm)
    for i in range(3):
        np.testing.assert_allclose(out[i].data, fields[i].smooth(sgm).data)

    out.data = 0
    fields.smooth(sigma=sgm, out=out)
    for i in range(3):
        np.testing.assert_allclose(out[i].data, fields[i].smooth(sgm).data)




"""
Solvers define how a pde is solved, i.e., advanced in time.

.. autosummary::
   :nosignatures:

   ~controller.Controller
   ~explicit.ExplicitSolver
   ~implicit.ImplicitSolver
   ~scipy.ScipySolver
   ~registered_solvers
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from typing import List

from .controller import Controller
from .explicit import ExplicitSolver
from .implicit import ImplicitSolver
from .scipy import ScipySolver



def registered_solvers() -> List[str]:
    """ returns all solvers that are registered in the package
    
    Returns:
        list of str: List with the names of the solvers
    """
    from .base import SolverBase
    return SolverBase.registered_solvers  # type: ignore



__all__ = ['Controller', 'ExplicitSolver', 'ImplicitSolver', 'ScipySolver',
           'registered_solvers']



r"""
This module implements differential operators on spherical grids 

.. autosummary::
   :nosignatures:

   make_laplace
   make_gradient
   make_divergence
   make_vector_gradient
   make_tensor_divergence
   
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from typing import Callable

import numpy as np
from scipy import sparse

from .common import make_general_poisson_solver
from .. import SphericalGrid
from ..boundaries import Boundaries
from ...tools.numba import jit_allocate_out
from ...tools.docstrings import fill_in_docstring



@fill_in_docstring
def make_laplace(bcs: Boundaries, conservative: bool = True) -> Callable:
    """ make a discretized laplace operator for a spherical grid
    
    {DESCR_SPHERICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        conservative (bool): flag indicating whether the laplace operator should
            be conservative (which results in slightly slower computations).    
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, SphericalGrid)
    bcs.check_value_rank(0)

    # calculate preliminary quantities
    dim_r = bcs.grid.shape[0]
    dr = bcs.grid.discretization[0]
    rs = bcs.grid.axes_coords[0]
    r_min, r_max = bcs.grid.axes_bounds[0]

    # prepare boundary values
    value_lower_bc = bcs[0].low.make_virtual_point_evaluator()
    value_upper_bc = bcs[0].high.make_virtual_point_evaluator()
    
    if conservative:
        # create a conservative spherical laplace operator
        rl = r_min + dr * np.arange(dim_r)  # inner radii of spherical shells
        rh = rl + dr                        # outer radii
        assert np.isclose(rh[-1], r_max)
        volumes = (rh**3 - rl**3) / 3       # volume of the spherical shells
    
        if r_min == 0:
            factors = dr / volumes
                    
            @jit_allocate_out(out_shape=(dim_r,))
            def laplace(arr, out=None):
                """ apply laplace operator to array `arr` """
                i = 0
                fp = (i + 1)**2 * (arr[i + 1] - arr[i])
                fm = 0
                out[i] = (fp - fm) * factors[i]
                
                for i in range(1, dim_r - 1):  # iterate inner radial points
                    fp = (i + 1)**2 * (arr[i + 1] - arr[i])
                    fm = i**2 * (arr[i] - arr[i - 1])
                    out[i] = (fp - fm) * factors[i]
                    
                # express boundary condition at outer side
                i = dim_r - 1
                arr_r_h = value_upper_bc(arr, (i,))
                fp = (i + 1)**2 * (arr_r_h - arr[i])
                fm = i**2 * (arr[i] - arr[i - 1])
                out[i] = (fp - fm) * factors[i]
                return out
            
        else:  # r_min > 0
            factor_l = (rs - 0.5 * dr)**2 / (dr * volumes)
            factor_h = (rs + 0.5 * dr)**2 / (dr * volumes)
            
            @jit_allocate_out(out_shape=(dim_r,))
            def laplace(arr, out=None):
                """ apply laplace operator to array `arr` """
                i = 0
                arr_r_l = value_lower_bc(arr, (i,))
                out[i] = (factor_h[i] * (arr[i + 1] - arr[i]) -
                          factor_l[i] * (arr[i] - arr_r_l))
                
                for i in range(1, dim_r - 1):  # iterate inner radial points
                    out[i] = (factor_h[i] * (arr[i + 1] - arr[i]) -
                              factor_l[i] * (arr[i] - arr[i - 1]))
                    
                # express boundary condition at outer side
                i = dim_r - 1
                arr_r_h = value_upper_bc(arr, (i,))
                out[i] = (factor_h[i] * (arr_r_h - arr[i]) -
                          factor_l[i] * (arr[i] - arr[i - 1]))
                return out

    else:  # not conservative
        dr2 = 1 / dr**2
    
        if r_min == 0:
            # create an operator that is not conservative
        
            @jit_allocate_out(out_shape=(dim_r,))
            def laplace(arr, out=None):
                """ apply laplace operator to array `arr` """
                i = 0
                out[i] = 3 * (arr[i + 1] - arr[i]) * dr2
                
                for i in range(1, dim_r - 1):  # iterate inner radial points
                    out[i] = ((arr[i + 1] - 2 * arr[i] + arr[i - 1]) * dr2 +
                              (arr[i + 1] - arr[i - 1]) / (i + 0.5) * dr2)
                    
                # express boundary condition at outer side
                i = dim_r - 1
                arr_r_h = value_upper_bc(arr, (i,))
                out[i] = ((arr_r_h - 2 * arr[i] + arr[i - 1]) * dr2 +
                          (arr_r_h - arr[i - 1]) / (i + 0.5) * dr2)
                return out
            
        else:  # r_min > 0
            @jit_allocate_out(out_shape=(dim_r,))
            def laplace(arr, out=None):
                """ apply laplace operator to array `arr` """
                i = 0
                arr_r_l = value_lower_bc(arr, (i,))
                out[i] = ((arr[i + 1] - 2 * arr[i] + arr_r_l) * dr2 +
                          (arr[i + 1] - arr_r_l) / (rs[i] * dr))
                
                for i in range(1, dim_r - 1):  # iterate inner radial points
                    out[i] = ((arr[i + 1] - 2 * arr[i] + arr[i - 1]) * dr2 +
                              (arr[i + 1] - arr[i - 1]) / (rs[i] * dr))
                    
                # express boundary condition at outer side
                i = dim_r - 1
                arr_r_h = value_upper_bc(arr, (i,))
                out[i] = ((arr_r_h - 2 * arr[i] + arr[i - 1]) * dr2 +
                          (arr_r_h - arr[i - 1]) / (rs[i] * dr))
                return out
        
    return laplace  # type: ignore



@fill_in_docstring
def make_gradient(bcs: Boundaries) -> Callable:
    """ make a discretized gradient operator for a spherical grid
    
    {DESCR_SPHERICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, SphericalGrid)
    bcs.check_value_rank(0)

    # calculate preliminary quantities
    dim_r = bcs.grid.shape[0]
    dr = bcs.grid.discretization[0]
    r_min, _ = bcs.grid.axes_bounds[0]
    
    scale_r = 1 / (2 * dr)
    
    # prepare boundary values
    boundary = bcs[0]
    value_lower_bc = boundary.low.make_virtual_point_evaluator()
    value_upper_bc = boundary.high.make_virtual_point_evaluator()
    
    @jit_allocate_out(out_shape=(3, dim_r))
    def gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        # no-flux at the origin 
        i = 0
        if r_min == 0:
            out[0, i] = (arr[1] - arr[0]) * scale_r
        else:
            arr_r_l = value_lower_bc(arr, (i,))
            out[0, i] = (arr[1] - arr_r_l) * scale_r            
        out[1, i] = out[2, i] = 0  # no angular dependence by definition
        
        for i in range(1, dim_r - 1):  # iterate inner radial points
            out[0, i] = (arr[i + 1] - arr[i - 1]) * scale_r
            out[1, i] = out[2, i] = 0  # no angular dependence by definition

        i = dim_r - 1
        arr_r_h = value_upper_bc(arr, (i,))
        out[0, i] = (arr_r_h - arr[i - 1]) * scale_r
        out[1, i] = out[2, i] = 0  # no angular dependence by definition
        
        return out
        
    return gradient  # type: ignore

    

@fill_in_docstring
def make_divergence(bcs: Boundaries) -> Callable:
    """ make a discretized divergence operator for a spherical grid
    
    {DESCR_SPHERICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, SphericalGrid)
    bcs.check_value_rank(0)

    # calculate preliminary quantities
    dim_r = bcs.grid.shape[0]
    dr = bcs.grid.discretization[0]
    rs = bcs.grid.axes_coords[0]
    r_min, _ = bcs.grid.axes_bounds[0]

    scale_r = 1 / (2 * dr)
    
    # prepare boundary values
    boundary = bcs[0]
    value_lower_bc = boundary.low.make_virtual_point_evaluator()
    value_upper_bc = boundary.high.make_virtual_point_evaluator()

    if r_min == 0:
        @jit_allocate_out(out_shape=(dim_r,))
        def divergence(arr, out=None):
            """ apply divergence operator to array `arr` """
            i = 0
            out[i] = (arr[0, 1] + 7 * arr[0, 0]) * scale_r
            
            for i in range(1, dim_r - 1):  # iterate inner radial points
                out[i] = (arr[0, i + 1] - arr[0, i - 1]) * scale_r + \
                         (2 * arr[0, i] / ((i + 0.5) * dr))
    
            i = dim_r - 1
            arr_r_h = value_upper_bc(arr[0], (i,))
            out[i] = (arr_r_h - arr[0, i - 1]) * scale_r + \
                     (2 * arr[0, i] / ((i + 0.5) * dr))
            
            return out
        
    else:  # r_min > 0
        fs = 2 / rs  # factors that need to be multiplied below
        
        @jit_allocate_out(out_shape=(dim_r,))
        def divergence(arr, out=None):
            """ apply divergence operator to array `arr` """            
            # inner radial boundary condition
            i = 0
            arr_r_l = value_lower_bc(arr[0], (i,))
            out[i] = (arr[0, i + 1] - arr_r_l) * scale_r + fs[i] * arr[0, i]
            
            for i in range(1, dim_r - 1):  # iterate radial points
                out[i] = ((arr[0, i + 1] - arr[0, i - 1]) * scale_r + 
                          fs[i] * arr[0, i])
                
            # outer radial boundary condition
            i = dim_r - 1
            arr_r_h = value_upper_bc(arr[0], (i,))
            out[i] = (arr_r_h - arr[0, i - 1]) * scale_r + fs[i] * arr[0, i]
                
            return out
                
    return divergence  # type: ignore


    
@fill_in_docstring
def make_vector_gradient(bcs: Boundaries) -> Callable:
    """ make a discretized vector gradient operator for a spherical grid
    
    {DESCR_SPHERICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, SphericalGrid)
    bcs.check_value_rank(1)

    gradient_r = make_gradient(bcs.extract_component(0))
    gradient_theta = make_gradient(bcs.extract_component(1))
    gradient_phi = make_gradient(bcs.extract_component(2))
        
    @jit_allocate_out(out_shape=(3, 3) + bcs.grid.shape)
    def vector_gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        gradient_r(arr[0], out=out[:, 0])
        gradient_theta(arr[1], out=out[:, 1])
        gradient_phi(arr[2], out=out[:, 2])
        return out    
        
    return vector_gradient  # type: ignore



@fill_in_docstring
def make_tensor_divergence(bcs: Boundaries) -> Callable:
    """ make a discretized tensor divergence operator for a spherical grid
    
    {DESCR_SPHERICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, SphericalGrid)
    bcs.check_value_rank(1)

    divergence_r = make_divergence(bcs.extract_component(0))
    divergence_theta = make_divergence(bcs.extract_component(1))
    divergence_phi = make_divergence(bcs.extract_component(2))
        
    @jit_allocate_out(out_shape=(3,) + bcs.grid.shape)
    def tensor_divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        divergence_r(arr[0], out=out[0])
        divergence_theta(arr[1], out=out[1])
        divergence_phi(arr[2], out=out[2])
        return out
        
    return tensor_divergence  # type: ignore



@fill_in_docstring
def _get_laplace_matrix(bcs):
    """ get sparse matrix for laplace operator on a polar grid
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        tuple: A sparse matrix and a sparse vector that can be used to evaluate
        the discretized laplacian
    """    
    assert isinstance(bcs.grid, SphericalGrid)
    bcs.check_value_rank(0)

    # calculate preliminary quantities
    dim_r = bcs.grid.shape[0]
    dr = bcs.grid.discretization[0]
    rs = bcs.grid.axes_coords[0]
    r_min, r_max = bcs.grid.axes_bounds[0]
    

    # create a conservative spherical laplace operator
    rl = r_min + dr * np.arange(dim_r)  # inner radii of spherical shells
    rh = rl + dr                        # outer radii
    assert np.isclose(rh[-1], r_max)
    volumes = (rh**3 - rl**3) / 3       # volume of the spherical shells

    factor_l = (rs - 0.5 * dr)**2 / (dr * volumes)
    factor_h = (rs + 0.5 * dr)**2 / (dr * volumes)

    matrix = sparse.dok_matrix((dim_r, dim_r))
    vector = sparse.dok_matrix((dim_r, 1))
    
    for i in range(dim_r):
        matrix[i, i] += -factor_l[i] - factor_h[i]
        
        if i == 0:
            if r_min == 0:
                matrix[i, i + 1] = factor_l[i]
            else:
                const, entries = bcs[0].get_data((-1,))
                vector[i] += const * factor_l[i]
                for k, v in entries.items():
                    matrix[i, k] += v * factor_l[i]

        else:
            matrix[i, i - 1] = factor_l[i]
            
        if i == dim_r - 1:
            const, entries = bcs[0].get_data((dim_r,))
            vector[i] += const * factor_h[i]
            for k, v in entries.items():
                matrix[i, k] += v * factor_h[i]
                
        else:
            matrix[i, i + 1] = factor_h[i]
            
    return matrix, vector



@fill_in_docstring
def make_poisson_solver(bcs: Boundaries, method: str = 'auto') -> Callable:
    """ make a operator that solves Poisson's equation
    
    {DESCR_POLAR_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    matrix, vector = _get_laplace_matrix(bcs)
    return make_general_poisson_solver(matrix, vector, method)


# register all operators with the grid class
SphericalGrid.register_operator('laplace', make_laplace,
                                rank_in=0, rank_out=0)
SphericalGrid.register_operator('gradient', make_gradient,
                                rank_in=0, rank_out=1)
SphericalGrid.register_operator('divergence', make_divergence,
                                rank_in=1, rank_out=0)
SphericalGrid.register_operator('vector_gradient', make_vector_gradient,
                                rank_in=1, rank_out=2)
SphericalGrid.register_operator('tensor_divergence', make_tensor_divergence,
                                rank_in=2, rank_out=1)
SphericalGrid.register_operator('poisson_solver', make_poisson_solver,
                                rank_in=0, rank_out=0)



'''
This module implements differential operators on Cartesian grids 

.. autosummary::
   :nosignatures:

   make_laplace
   make_gradient
   make_divergence
   make_vector_gradient
   make_vector_laplace
   make_tensor_divergence
   
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>   
'''

from typing import Callable, Tuple, Any

import numba as nb
import numpy as np
from scipy import ndimage, sparse

from .common import (make_laplace_from_matrix, make_general_poisson_solver,
                     PARALLELIZATION_THRESHOLD_2D, PARALLELIZATION_THRESHOLD_3D)
from ..boundaries import Boundaries
from ..cartesian import CartesianGridBase
from ...tools.numba import jit_allocate_out
from ...tools.docstrings import fill_in_docstring



@fill_in_docstring
def _get_laplace_matrix_1d(bcs) -> Tuple[Any, Any]:
    """ get sparse matrix for laplace operator on a 1d Cartesian grid
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        tuple: A sparse matrix and a sparse vector that can be used to evaluate
        the discretized laplacian
    """    
    dim_x = bcs.grid.shape[0]
    matrix = sparse.dok_matrix((dim_x, dim_x))
    vector = sparse.dok_matrix((dim_x, 1))
    
    for i in range(dim_x):
        matrix[i, i] += -2
        
        if i == 0:
            const, entries = bcs[0].get_data((-1,))
            vector[i] += const
            for k, v in entries.items():
                matrix[i, k] += v
        else:
            matrix[i, i - 1] += 1
            
        if i == dim_x - 1:
            const, entries = bcs[0].get_data((dim_x,))
            vector[i] += const
            for k, v in entries.items():
                matrix[i, k] += v
        else:
            matrix[i, i + 1] += 1

    matrix *= bcs.grid.discretization[0]**-2
    vector *= bcs.grid.discretization[0]**-2
            
    return matrix, vector

    

@fill_in_docstring
def _get_laplace_matrix_2d(bcs) -> Tuple[Any, Any]:
    """ get sparse matrix for laplace operator on a 2d Cartesian grid
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        tuple: A sparse matrix and a sparse vector that can be used to evaluate
        the discretized laplacian
    """        
    dim_x, dim_y = bcs.grid.shape
    matrix = sparse.dok_matrix((dim_x * dim_y, dim_x * dim_y))
    vector = sparse.dok_matrix((dim_x * dim_y, 1))
    
    bc_x, bc_y = bcs
    scale_x, scale_y = bcs.grid.discretization**-2
    
    def i(x, y):
        """ helper function for flattening the index
        
        This is equivalent to np.ravel_multi_index((x, y), (dim_x, dim_y))
        """
        return x * dim_y + y
        
    # set diagonal elements, i.e., the central value in the kernel
    matrix.setdiag(-2 * (scale_x + scale_y))
        
    for x in range(dim_x):
        for y in range(dim_y):
            # handle x-direction
            if x == 0:
                const, entries = bc_x.get_data((-1, y))
                vector[i(x, y)] += const * scale_x
                for k, v in entries.items():
                    matrix[i(x, y), i(k, y)] += v * scale_x
            else:
                matrix[i(x, y), i(x - 1, y)] += scale_x

            if x == dim_x - 1:
                const, entries = bc_x.get_data((dim_x, y))
                vector[i(x, y)] += const * scale_x
                for k, v in entries.items():
                    matrix[i(x, y), i(k, y)] += v * scale_x
            else:
                matrix[i(x, y), i(x + 1, y)] += scale_x  
                
            # handle y-direction
            if y == 0:
                const, entries = bc_y.get_data((x, -1))
                vector[i(x, y)] += const * scale_y
                for k, v in entries.items():
                    matrix[i(x, y), i(x, k)] += v * scale_y
            else:
                matrix[i(x, y), i(x, y - 1)] += scale_y

            if y == dim_y - 1:
                const, entries = bc_y.get_data((x, dim_y))
                vector[i(x, y)] += const * scale_y
                for k, v in entries.items():
                    matrix[i(x, y), i(x, k)] += v * scale_y
            else:
                matrix[i(x, y), i(x, y + 1)] += scale_y

    return matrix, vector



@fill_in_docstring
def _get_laplace_matrix(bcs: Boundaries):
    """ get sparse matrix for laplace operator on a 1d Cartesian grid
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        tuple: A sparse matrix and a sparse vector that can be used to evaluate
        the discretized laplacian
    """    
    dim = bcs.grid.dim
    bcs.check_value_rank(0)
    
    if dim == 1:
        result = _get_laplace_matrix_1d(bcs)
    elif dim == 2:
        result = _get_laplace_matrix_2d(bcs)
    else:
        raise NotImplementedError('Numba laplace operator not implemented '
                                  f'for {dim:d} dimensions')
                                      
    return result



def _make_derivative(bcs: Boundaries, axis: int = 0, method: str = 'central') \
        -> Callable:
    """ make a derivative operator along a single axis using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        axis (int):
            The axis along which the derivative will be taken
        method (str):
            The method for calculating the derivative. Possible values are
            'central', 'forward', and 'backward'.
        
    Returns:
        A function that can be applied to an array of values. The result will be
        an array of the same shape containing the actual derivatives at the grid
        points.
    """
    if method not in {'central', 'forward', 'backward'}:
        raise ValueError(f'Unknown derivative type `{method}`')
    
    shape = bcs.grid.shape
    dim = len(shape)
    dx = bcs.grid.discretization[axis]
    region = bcs[axis].make_region_evaluator()
        
    if dim == 1:
        @jit_allocate_out(out_shape=shape)
        def diff(arr, out=None):
            """ calculate derivative of 1d array `arr` """
            for i in range(shape[0]):
                arr_l, arr_m, arr_h = region(arr, (i,))
                if method == 'central':
                    out[i] = (arr_h - arr_l) * 0.5 / dx
                elif method == 'forward':
                    out[i] = (arr_h - arr_m) / dx
                elif method == 'backward':
                    out[i] = (arr_m - arr_l) / dx
                    
            return out       
        
    elif dim == 2:
        @jit_allocate_out(out_shape=shape)
        def diff(arr, out=None):
            """ calculate derivative of 2d array `arr` """
            for i in range(shape[0]):
                for j in range(shape[1]):
                    arr_l, arr_m, arr_h = region(arr, (i, j))
                    if method == 'central':
                        out[i, j] = (arr_h - arr_l) * 0.5 / dx
                    elif method == 'forward':
                        out[i, j] = (arr_h - arr_m) / dx
                    elif method == 'backward':
                        out[i, j] = (arr_m - arr_l) / dx
                    
            return out       
        
    elif dim == 3:
        @jit_allocate_out(out_shape=shape)
        def diff(arr, out=None):
            """ calculate derivative of 3d array `arr` """
            for i in range(shape[0]):
                for j in range(shape[1]):
                    for k in range(shape[2]):
                        arr_l, arr_m, arr_h = region(arr, (i, j, k))
                        if method == 'central':
                            out[i, j, k] = (arr_h - arr_l) * 0.5 / dx
                        elif method == 'forward':
                            out[i, j, k] = (arr_h - arr_m) / dx
                        elif method == 'backward':
                            out[i, j, k] = (arr_m - arr_l) / dx
                    
            return out     
        
    else:
        raise NotImplementedError('Numba derivative operator not implemented '
                                  f'for {dim:d} dimensions')
        
    return diff  # type: ignore    



@fill_in_docstring
def _make_laplace_scipy_nd(bcs: Boundaries) -> Callable:
    """ make a laplace operator using the scipy module
    
    This only supports uniform discretizations.
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    scaling = bcs._uniform_discretization**-2
    args = bcs._scipy_border_mode
    
    def laplace(arr, out=None):
        """ apply laplace operator to array `arr` """
        return ndimage.laplace(scaling * arr, output=out, **args)
        
    return laplace


# def _make_laplace_fft_nd(shape, boundaries, dx=1):
#     """ make a laplace operator using the fft module """
#     
#     if not boundaries.periodic:
#         raise ValueError('Boundary needs to be periodic')
#     
#     dim = len(shape)
#     dx = np.broadcast_to(dx, (dim,))
#     ks = [np.fft.fftfreq(n, d) for n, d in zip(shape, dx)]
#     k2s = np.meshgrid(*ks, indexing='ij')**2
#     
#     def laplace(arr, out=None):
#         """ apply laplace operator to array `arr` """
#         arr_k = np.fft.rfftn(arr, s=shape)
#         return np.fft.irfftn(arr_k * k2s, s=shape)
#         
#     return laplace

        
    
@fill_in_docstring
def _make_laplace_numba_1d(bcs: Boundaries) -> Callable:
    """ make a 1d laplace operator using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values    
    """
    dim_x = bcs.grid.shape[0]
    scale = bcs.grid.discretization[0]**-2
    region_x = bcs[0].make_region_evaluator()
    
    @jit_allocate_out
    def laplace(arr, out=None):
        """ apply laplace operator to array `arr` """
        for i in range(dim_x):
            valm, val, valp = region_x(arr, (i,))
            out[i] = (valm - 2 * val + valp) * scale

        return out        
        
    return laplace  # type: ignore

    
    
@fill_in_docstring
def _make_laplace_numba_2d(bcs: Boundaries) -> Callable:
    """ make a 2d laplace operator using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values    
    """
    dim_x, dim_y = bcs.grid.shape
    scale_x, scale_y = bcs.grid.discretization**-2
    
    region_x = bcs[0].make_region_evaluator()
    region_y = bcs[1].make_region_evaluator()
    
    # use parallel processing for large enough arrays 
    parallel = (dim_x * dim_y >= PARALLELIZATION_THRESHOLD_2D**2)
    
    @jit_allocate_out(parallel=parallel)
    def laplace(arr, out=None):
        """ apply laplace operator to array `arr` """
        for i in nb.prange(dim_x):

            # left-most part
            j = 0
            arr_x_l, arr_c, arr_x_h = region_x(arr, (i, j))
            arr_y_l, _, arr_y_h = region_y(arr, (i, j))
            lap_x = (arr_x_l - 2 * arr_c + arr_x_h) * scale_x
            lap_y = (arr_y_l - 2 * arr_c + arr_y_h) * scale_y
            out[i, j] = lap_x + lap_y
            
            if dim_y == 1:
                continue  # deal with singular y dimension
 
            # middle part without boundary conditions in y-direction
            for j in range(1, dim_y - 1):
                arr_x_l, arr_c, arr_x_h = region_x(arr, (i, j))
                arr_y_l = arr[i, j - 1]
                arr_y_h = arr[i, j + 1]
                 
                lap_x = (arr_x_l - 2 * arr_c + arr_x_h) * scale_x
                lap_y = (arr_y_l - 2 * arr_c + arr_y_h) * scale_y
                out[i, j] = lap_x + lap_y
         
            # right-most part
            j = dim_y - 1
            arr_x_l, arr_c, arr_x_h = region_x(arr, (i, j))
            arr_y_l, _, arr_y_h = region_y(arr, (i, j))
            lap_x = (arr_x_l - 2 * arr_c + arr_x_h) * scale_x
            lap_y = (arr_y_l - 2 * arr_c + arr_y_h) * scale_y
            out[i, j] = lap_x + lap_y

        return out        
        
    return laplace  # type: ignore
    


@fill_in_docstring
def _make_laplace_numba_3d(bcs: Boundaries) -> Callable:
    """ make a 3d laplace operator using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    dim_x, dim_y, dim_z = bcs.grid.shape 
    scale_x, scale_y, scale_z = bcs.grid.discretization**-2
    
    region_x = bcs[0].make_region_evaluator()
    region_y = bcs[1].make_region_evaluator()
    region_z = bcs[2].make_region_evaluator()
    
    # use parallel processing for large enough arrays 
    parallel = (dim_x * dim_y * dim_z >= PARALLELIZATION_THRESHOLD_3D**3)
    
    @jit_allocate_out(parallel=parallel)
    def laplace(arr, out=None):
        """ apply laplace operator to array `arr` """
        for i in nb.prange(dim_x):
            for j in range(dim_y):
                for k in range(dim_z):

                    arr_x_l, arr_c, arr_x_h = region_x(arr, (i, j, k))
                    arr_y_l, _, arr_y_h = region_y(arr, (i, j, k))
                    arr_z_l, _, arr_z_h = region_z(arr, (i, j, k))
                    
                    lap_x = (arr_x_l - 2 * arr_c + arr_x_h) * scale_x
                    lap_y = (arr_y_l - 2 * arr_c + arr_y_h) * scale_y
                    lap_z = (arr_z_l - 2 * arr_c + arr_z_h) * scale_z
                    out[i, j, k] = lap_x + lap_y + lap_z

        return out        
        
    return laplace  # type: ignore



@fill_in_docstring
def make_laplace(bcs: Boundaries, method: str = 'auto') -> Callable:
    """ make a laplace operator on a Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        method (str): Method used for calculating the laplace operator.
            If method='auto', a suitable method is chosen automatically.
        
    Returns:
        A function that can be applied to an array of values
    """
    dim = bcs.grid.dim
    bcs.check_value_rank(0)
    
    if method == 'auto':
        # choose the fastest available Laplace operator
        if 1 <= dim <= 3:
            method = 'numba'
        else:
            method = 'scipy'
            
    if method == 'numba':
        if dim == 1:
            laplace = _make_laplace_numba_1d(bcs)
        elif dim == 2:
            laplace = _make_laplace_numba_2d(bcs)
        elif dim == 3:
            laplace = _make_laplace_numba_3d(bcs)
        else:
            raise NotImplementedError('Numba laplace operator not implemented '
                                      f'for {dim:d} dimensions')
                          
    elif method == 'matrix':
        laplace = make_laplace_from_matrix(*_get_laplace_matrix(bcs))
                                      
    elif method == 'scipy':
        laplace = _make_laplace_scipy_nd(bcs)
        
    else:
        raise ValueError(f'Method `{method}` is not defined')
        
    return laplace


# register operators with the grid class
CartesianGridBase.register_operator('laplace', make_laplace,
                                    rank_in=0, rank_out=0)



@fill_in_docstring
def _make_gradient_scipy_nd(bcs: Boundaries) -> Callable:
    """ make a gradient operator using the scipy module
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """    
    scaling = 0.5 / bcs._uniform_discretization
    dim = bcs.grid.dim
    shape_out = (dim,) + bcs.grid.shape 
    args = bcs._scipy_border_mode
    
    def gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        if out is None:
            out = np.empty(shape_out)
        for i in range(dim):
            out[i] = ndimage.convolve1d(arr, [1, 0, -1], axis=i, **args) 
        return out * scaling
        
    return gradient



@fill_in_docstring
def _make_gradient_numba_1d(bcs: Boundaries) -> Callable:
    """ make a 1d gradient operator using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:    
        A function that can be applied to an array of values
    """
    dim_x = bcs.grid.shape[0]
    scale = 0.5 / bcs.grid.discretization[0]
    region_x = bcs[0].make_region_evaluator()
    
    @jit_allocate_out(out_shape=(1, dim_x))
    def gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        for i in range(dim_x):
            valm, _, valp = region_x(arr, (i,))
            out[0, i] = (valp - valm) * scale
                
        return out        
        
    return gradient  # type: ignore



@fill_in_docstring
def _make_gradient_numba_2d(bcs: Boundaries) -> Callable:
    """ make a 2d gradient operator using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values    
    """
    dim_x, dim_y = bcs.grid.shape
    scale_x, scale_y = 0.5 / bcs.grid.discretization
    
    region_x = bcs[0].make_region_evaluator()
    region_y = bcs[1].make_region_evaluator()

    # use parallel processing for large enough arrays 
    parallel = (dim_x * dim_y >= PARALLELIZATION_THRESHOLD_2D**2)
        
    @jit_allocate_out(parallel=parallel, out_shape=(2, dim_x, dim_y))
    def gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        for i in nb.prange(dim_x):
            for j in range(dim_y):
                arr_x_l, _, arr_x_h = region_x(arr, (i, j))
                arr_y_l, _, arr_y_h = region_y(arr, (i, j))
                    
                out[0, i, j] = (arr_x_h - arr_x_l) * scale_x
                out[1, i, j] = (arr_y_h - arr_y_l) * scale_y
                
        return out        
        
    return gradient  # type: ignore



@fill_in_docstring
def _make_gradient_numba_3d(bcs: Boundaries) -> Callable:
    """ make a 3d gradient operator using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values    
    """
    dim_x, dim_y, dim_z = bcs.grid.shape
    scale_x, scale_y, scale_z = 0.5 / bcs.grid.discretization
    
    region_x = bcs[0].make_region_evaluator()
    region_y = bcs[1].make_region_evaluator()
    region_z = bcs[2].make_region_evaluator()
    
    # use parallel processing for large enough arrays 
    parallel = (dim_x * dim_y * dim_z >= PARALLELIZATION_THRESHOLD_3D**3)
    
    @jit_allocate_out(parallel=parallel, out_shape=(3, dim_x, dim_y, dim_z))
    def gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        for i in nb.prange(dim_x):
            for j in range(dim_y):
                for k in range(dim_z):
                    arr_x_l, _, arr_x_h = region_x(arr, (i, j, k))
                    arr_y_l, _, arr_y_h = region_y(arr, (i, j, k))
                    arr_z_l, _, arr_z_h = region_z(arr, (i, j, k))

                    out[0, i, j, k] = (arr_x_h - arr_x_l) * scale_x
                    out[1, i, j, k] = (arr_y_h - arr_y_l) * scale_y
                    out[2, i, j, k] = (arr_z_h - arr_z_l) * scale_z
                
        return out        
        
    return gradient  # type: ignore



@fill_in_docstring
def make_gradient(bcs: Boundaries, method: str = 'auto') -> Callable:
    """ make a gradient operator on a Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        method (str): Method used for calculating the gradient operator.
            If method='auto', a suitable method is chosen automatically.
        
    Returns:
        A function that can be applied to an array of values
    """
    dim = bcs.grid.dim
    bcs.check_value_rank(0)
    
    if method == 'auto':
        # choose the fastest available gradient operator
        if 1 <= dim <= 3:
            method = 'numba'
        else:
            method = 'scipy'

    if method == 'numba':
        if dim == 1:
            gradient = _make_gradient_numba_1d(bcs)
        elif dim == 2:
            gradient = _make_gradient_numba_2d(bcs)
        elif dim == 3:
            gradient = _make_gradient_numba_3d(bcs)
        else:
            raise NotImplementedError('Numba gradient operator not '
                                      f'implemented for {dim}')
                                      
    elif method == 'scipy':
        gradient = _make_gradient_scipy_nd(bcs)
        
    else:
        raise ValueError(f'Method `{method}` is not defined')
        
    return gradient


# register operators with the grid class
CartesianGridBase.register_operator('gradient', make_gradient,
                                    rank_in=0, rank_out=1)



@fill_in_docstring
def _make_divergence_scipy_nd(bcs: Boundaries) -> Callable:
    """ make a divergence operator using the scipy module
    
    This only supports uniform discretizations.
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    shape = bcs.grid.shape
    scaling = 0.5 / bcs._uniform_discretization
    args = bcs._scipy_border_mode
    
    def divergence(arr, out=None):
        """ apply divergence operator to array `arr` """
        assert arr.shape[0] == len(shape) and len(arr.shape) == len(shape) + 1
        # need to initialize with zeros since data is added later
        if out is None:
            out = np.zeros(arr.shape[1:])
        else:
            out[:] = 0
            
        for i in range(len(shape)):
            out += ndimage.convolve1d(arr[i], [1, 0, -1], axis=i, **args) 
        return out * scaling
        
    return divergence



@fill_in_docstring
def _make_divergence_numba_1d(bcs: Boundaries) -> Callable:
    """ make a 1d divergence operator using numba compilation
    
    Args:
        dim (int): The number of support points for each axes
        boundaries (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        dx (float): The discretization
        
    Returns:    
        A function that can be applied to an array of values
    """
    dim_x = bcs.grid.shape[0]
    scale = 0.5 / bcs.grid.discretization[0]
    region_x = bcs[0].make_region_evaluator()
    
    @jit_allocate_out(out_shape=(dim_x,))
    def divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        for i in range(dim_x):
            valm, _, valp = region_x(arr[0], (i,))
            out[i] = (valp - valm) * scale
                
        return out        
        
    return divergence  # type: ignore



@fill_in_docstring
def _make_divergence_numba_2d(bcs: Boundaries) -> Callable:
    """ make a 2d divergence operator using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values    
    """
    dim_x, dim_y = bcs.grid.shape
    scale_x, scale_y = 0.5 / bcs.grid.discretization
    
    region_x = bcs[0].make_region_evaluator()
    region_y = bcs[1].make_region_evaluator()

    # use parallel processing for large enough arrays 
    parallel = (dim_x * dim_y >= PARALLELIZATION_THRESHOLD_2D**2)
    
    @jit_allocate_out(parallel=parallel, out_shape=(dim_x, dim_y))
    def divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        for i in nb.prange(dim_x):
            for j in range(dim_y):
                arr_x_l, _, arr_x_h = region_x(arr[0], (i, j))
                arr_y_l, _, arr_y_h = region_y(arr[1], (i, j))
                    
                d_x = (arr_x_h - arr_x_l) * scale_x
                d_y = (arr_y_h - arr_y_l) * scale_y
                out[i, j] = d_x + d_y
                
        return out        
        
    return divergence  # type: ignore



@fill_in_docstring
def _make_divergence_numba_3d(bcs: Boundaries) -> Callable:
    """ make a 3d divergence operator using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values    
    """
    dim_x, dim_y, dim_z = bcs.grid.shape
    scale_x, scale_y, scale_z = 0.5 / bcs.grid.discretization
    
    region_x = bcs[0].make_region_evaluator()
    region_y = bcs[1].make_region_evaluator()
    region_z = bcs[2].make_region_evaluator()
    
    # use parallel processing for large enough arrays 
    parallel = (dim_x * dim_y * dim_z >= PARALLELIZATION_THRESHOLD_3D**3)
    
    @jit_allocate_out(parallel=parallel, out_shape=(dim_x, dim_y, dim_z))
    def divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        for i in nb.prange(dim_x):
            for j in range(dim_y):
                for k in range(dim_z):
                    arr_x_l, _, arr_x_h = region_x(arr[0], (i, j, k))
                    arr_y_l, _, arr_y_h = region_y(arr[1], (i, j, k))
                    arr_z_l, _, arr_z_h = region_z(arr[2], (i, j, k))
                        
                    d_x = (arr_x_h - arr_x_l) * scale_x
                    d_y = (arr_y_h - arr_y_l) * scale_y
                    d_z = (arr_z_h - arr_z_l) * scale_z
                    out[i, j, k] = d_x + d_y + d_z
                
        return out        
        
    return divergence  # type: ignore



@fill_in_docstring
def make_divergence(bcs: Boundaries, method: str = 'auto') -> Callable:
    """ make a divergence operator on a Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        method (str): Method used for calculating the divergence operator.
            If method='auto', a suitable method is chosen automatically.
        
    Returns:
        A function that can be applied to an array of values
    """    
    dim = bcs.grid.dim
    bcs.check_value_rank(0)
    
    if method == 'auto':
        # choose the fastest available divergence operator
        if 1 <= dim <= 3:
            method = 'numba'
        else:
            method = 'scipy'

    if method == 'numba':
        if dim == 1:
            divergence = _make_divergence_numba_1d(bcs)
        elif dim == 2:
            divergence = _make_divergence_numba_2d(bcs)
        elif dim == 3:
            divergence = _make_divergence_numba_3d(bcs)
        else:
            raise NotImplementedError('Numba divergence operator not '
                                      f'implemented for {dim}')
                                      
    elif method == 'scipy':
        divergence = _make_divergence_scipy_nd(bcs)
        
    else:
        raise ValueError(f'Method `{method}` is not defined')
        
    return divergence


# register operators with the grid class
CartesianGridBase.register_operator('divergence', make_divergence,
                                    rank_in=1, rank_out=0)



@fill_in_docstring
def _make_vector_gradient_scipy_nd(bcs: Boundaries) -> Callable:
    """ make a vector gradient operator using the scipy module
    
    This only supports uniform discretizations.
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    scaling = 0.5 / bcs._uniform_discretization
    args = bcs._scipy_border_mode
    dim = bcs.grid.dim
    shape_out = (dim, dim) + bcs.grid.shape 
    
    def vector_gradient(arr, out=None):
        """ apply vector gradient operator to array `arr` """
        if out is None:
            out = np.empty(shape_out)
        for i in range(dim):
            for j in range(dim):
                out[i, j] = ndimage.convolve1d(arr[j], [1, 0, -1],
                                               axis=i, **args) 
        return out * scaling
        
    return vector_gradient



@fill_in_docstring
def _make_vector_gradient_numba_1d(bcs: Boundaries) -> Callable:
    """ make a 1d vector gradient operator using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:    
        A function that can be applied to an array of values
    """
    gradient = _make_gradient_numba_1d(bcs.extract_component(0))
        
    @jit_allocate_out(out_shape=(1, 1) + bcs.grid.shape)
    def vector_gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        gradient(arr[0], out=out[0])
        return out
        
    return vector_gradient  # type: ignore



@fill_in_docstring
def _make_vector_gradient_numba_2d(bcs: Boundaries) -> Callable:
    """ make a 2d vector gradient operator using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values    
    """
    gradient_x = _make_gradient_numba_2d(bcs.extract_component(0))
    gradient_y = _make_gradient_numba_2d(bcs.extract_component(1))
        
    @jit_allocate_out(out_shape=(2, 2) + bcs.grid.shape)
    def vector_gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        gradient_x(arr[0], out=out[:, 0])
        gradient_y(arr[1], out=out[:, 1])
        return out        
        
    return vector_gradient  # type: ignore



@fill_in_docstring
def _make_vector_gradient_numba_3d(bcs: Boundaries) -> Callable:
    """ make a 3d vector gradient operator using numba compilation
    
    Args:
        shape (tuple): The number of support points for each axes
        boundaries (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        dx (float or list): The discretizations
        
    Returns:
        A function that can be applied to an array of values    
    """
    gradient_x = _make_gradient_numba_3d(bcs.extract_component(0))
    gradient_y = _make_gradient_numba_3d(bcs.extract_component(1))
    gradient_z = _make_gradient_numba_3d(bcs.extract_component(2))
        
    @jit_allocate_out(out_shape=(3, 3) + bcs.grid.shape)
    def vector_gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        gradient_x(arr[0], out=out[:, 0])
        gradient_y(arr[1], out=out[:, 1])
        gradient_z(arr[2], out=out[:, 2])
        return out        
        
    return vector_gradient  # type: ignore



@fill_in_docstring
def make_vector_gradient(bcs: Boundaries, method: str = 'auto') -> Callable:
    """ make a vector gradient operator on a Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        method (str): Method used for calculating the vector gradient operator.
            If method='auto', a suitable method is chosen automatically
        
    Returns:
        A function that can be applied to an array of values
    """
    dim = bcs.grid.dim
    bcs.check_value_rank(1)
    
    # choose the fastest available vector gradient operator
    if method == 'auto':
        if 1 <= dim <= 3:
            method = 'numba'
        else:
            method = 'scipy'

    if method == 'numba':
        if dim == 1:
            gradient = _make_vector_gradient_numba_1d(bcs)
        elif dim == 2:
            gradient = _make_vector_gradient_numba_2d(bcs)
        elif dim == 3:
            gradient = _make_vector_gradient_numba_3d(bcs)
        else:
            raise NotImplementedError('Numba vector gradient operator not '
                                      f'implemented for {dim}')
                                      
    elif method == 'scipy':
        gradient = _make_vector_gradient_scipy_nd(bcs)
    else:
        raise ValueError(f'Method `{method}` is not defined')
        
    return gradient


# register operators with the grid class
CartesianGridBase.register_operator('vector_gradient', make_vector_gradient,
                                    rank_in=1, rank_out=2)



@fill_in_docstring
def _make_vector_laplace_scipy_nd(bcs: Boundaries) -> Callable:
    """ make a vector Laplacian using the scipy module
    
    This only supports uniform discretizations.
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    scaling = bcs._uniform_discretization**-2
    args = bcs._scipy_border_mode
    dim = bcs.grid.dim
    shape_out = (dim,) + bcs.grid.shape
    
    def vector_laplace(arr, out=None):
        """ apply vector Laplacian operator to array `arr` """
        if out is None:
            out = np.empty(shape_out)
        for i in range(dim):
            ndimage.laplace(arr[i], output=out[i], **args)
        return out * scaling
        
    return vector_laplace



@fill_in_docstring
def _make_vector_laplace_numba_1d(bcs: Boundaries) -> Callable:
    """ make a 1d vector Laplacian using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:    
        A function that can be applied to an array of values
    """
    laplace = _make_laplace_numba_1d(bcs.extract_component(0))
        
    @jit_allocate_out(out_shape=(1,) + bcs.grid.shape)
    def vector_laplace(arr, out=None):
        """ apply vector Laplacian to array `arr` """
        laplace(arr[0], out=out[0])
        return out
        
    return vector_laplace  # type: ignore



@fill_in_docstring
def _make_vector_laplace_numba_2d(bcs: Boundaries) -> Callable:
    """ make a 2d vector Laplacian using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values    
    """
    laplace_x = _make_laplace_numba_2d(bcs.extract_component(0))
    laplace_y = _make_laplace_numba_2d(bcs.extract_component(1))
        
    @jit_allocate_out(out_shape=(2,) + bcs.grid.shape)
    def vector_laplace(arr, out=None):
        """ apply vector Laplacian  to array `arr` """
        laplace_x(arr[0], out=out[0])
        laplace_y(arr[1], out=out[1])
        return out        
        
    return vector_laplace  # type: ignore



@fill_in_docstring
def _make_vector_laplace_numba_3d(bcs: Boundaries) -> Callable:
    """ make a 3d vector Laplacian using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values    
    """
    laplace_x = _make_laplace_numba_3d(bcs.extract_component(0))
    laplace_y = _make_laplace_numba_3d(bcs.extract_component(1))
    laplace_z = _make_laplace_numba_3d(bcs.extract_component(2))
        
    @jit_allocate_out(out_shape=(3,) + bcs.grid.shape)
    def vector_laplace(arr, out=None):
        """ apply vector Laplacian to array `arr` """
        laplace_x(arr[0], out=out[0])
        laplace_y(arr[1], out=out[1])
        laplace_z(arr[2], out=out[2])
        return out        
        
    return vector_laplace  # type: ignore



@fill_in_docstring
def make_vector_laplace(bcs: Boundaries, method: str = 'auto') -> Callable:
    """ make a vector Laplacian on a Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        method (str): Method used for calculating the vector laplace operator.
            If method='auto', a suitable method is chosen automatically.
        
    Returns:
        A function that can be applied to an array of values
    """
    dim = bcs.grid.dim
    bcs.check_value_rank(1)
    
    # choose the fastest available vector gradient operator
    if method == 'auto':
        if 1 <= dim <= 3:
            method = 'numba'
        else:
            method = 'scipy'

    if method == 'numba':
        if dim == 1:
            gradient = _make_vector_laplace_numba_1d(bcs)
        elif dim == 2:
            gradient = _make_vector_laplace_numba_2d(bcs)
        elif dim == 3:
            gradient = _make_vector_laplace_numba_3d(bcs)
        else:
            raise NotImplementedError('Numba vector gradient operator not '
                                      f'implemented for {dim}')
                                      
    elif method == 'scipy':
        gradient = _make_vector_laplace_scipy_nd(bcs)
    else:
        raise ValueError(f'Method `{method}` is not defined')
        
    return gradient


# register operators with the grid class
CartesianGridBase.register_operator('vector_laplace', make_vector_laplace,
                                    rank_in=1, rank_out=1)



@fill_in_docstring
def _make_tensor_divergence_scipy_nd(bcs: Boundaries) -> Callable:
    """ make a tensor divergence operator using the scipy module
    
    This only supports uniform discretizations.
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    scaling = 0.5 / bcs._uniform_discretization
    args = bcs._scipy_border_mode
    dim = bcs.grid.dim
    shape_out = (dim,) + bcs.grid.shape
    
    def tensor_divergence(arr, out=None):
        """ apply tensor divergence operator to array `arr` """
        # need to initialize with zeros since data is added later
        if out is None:
            out = np.zeros(shape_out)
        else:
            out[:] = 0
            
        for i in range(dim):
            for j in range(dim):
                out[i] += ndimage.convolve1d(arr[i, j], [1, 0, -1],
                                             axis=j, **args) 
        return out * scaling
        
    return tensor_divergence



@fill_in_docstring
def _make_tensor_divergence_numba_1d(bcs: Boundaries) -> Callable:
    """ make a 1d tensor divergence  operator using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:    
        A function that can be applied to an array of values
    """
    divergence = _make_divergence_numba_1d(bcs.extract_component(0))
        
    @jit_allocate_out(out_shape=(1,) + bcs.grid.shape)
    def tensor_divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        divergence(arr[0], out=out[0])
        return out
        
    return tensor_divergence  # type: ignore



@fill_in_docstring
def _make_tensor_divergence_numba_2d(bcs: Boundaries) -> Callable:
    """ make a 2d tensor divergence  operator using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values    
    """
    divergence_x = _make_divergence_numba_2d(bcs.extract_component(0))
    divergence_y = _make_divergence_numba_2d(bcs.extract_component(1))
        
    @jit_allocate_out(out_shape=(2,) + bcs.grid.shape)
    def tensor_divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        divergence_x(arr[0], out=out[0])
        divergence_y(arr[1], out=out[1])
        return out        
        
    return tensor_divergence  # type: ignore



@fill_in_docstring
def _make_tensor_divergence_numba_3d(bcs: Boundaries) -> Callable:
    """ make a 3d tensor divergence  operator using numba compilation
    
    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values    
    """
    divergence_x = _make_divergence_numba_3d(bcs.extract_component(0))
    divergence_y = _make_divergence_numba_3d(bcs.extract_component(1))
    divergence_z = _make_divergence_numba_3d(bcs.extract_component(2))
        
    @jit_allocate_out(out_shape=(3,) + bcs.grid.shape)
    def tensor_divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        divergence_x(arr[0], out=out[0])
        divergence_y(arr[1], out=out[1])
        divergence_z(arr[2], out=out[2])
        return out        
        
    return tensor_divergence  # type: ignore



@fill_in_docstring
def make_tensor_divergence(bcs: Boundaries, method: str = 'auto') -> Callable:
    """ make a tensor divergence operator on a Cartesian grid

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        method (str): Method used for calculating the tensor divergence
            operator. If method='auto', a suitable method is chosen
            automatically.
        
    Returns:
        A function that can be applied to an array of values
    """
    dim = bcs.grid.dim
    bcs.check_value_rank(1)
    
    # choose the fastest available tensor divergence operator
    if method == 'auto':
        if 1 <= dim <= 3:
            method = 'numba'
        else:
            method = 'scipy'

    if method == 'numba':
        if dim == 1:
            func = _make_tensor_divergence_numba_1d(bcs)
        elif dim == 2:
            func = _make_tensor_divergence_numba_2d(bcs)
        elif dim == 3:
            func = _make_tensor_divergence_numba_3d(bcs)
        else:
            raise NotImplementedError('Numba tensor divergence operator not '
                                      f'implemented for {dim}')
                                      
    elif method == 'scipy':
        func = _make_tensor_divergence_scipy_nd(bcs)
    else:
        raise ValueError(f'Method `{method}` is not defined')
        
    return func


# register operators with the grid class
CartesianGridBase.register_operator('tensor_divergence', make_tensor_divergence,
                                    rank_in=2, rank_out=1)



@fill_in_docstring
def make_poisson_solver(bcs: Boundaries, method: str = 'auto') -> Callable:
    """ make a operator that solves Poisson's equation

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        method (str): Method used for calculating the tensor divergence
            operator. If method='auto', a suitable method is chosen
            automatically.
        
    Returns:
        A function that can be applied to an array of values
    """
    matrix, vector = _get_laplace_matrix(bcs)
    return make_general_poisson_solver(matrix, vector, method)


# register operators with the grid class
CartesianGridBase.register_operator('poisson_solver', make_poisson_solver,
                                    rank_in=0, rank_out=0)



__all__ = ["make_laplace", "make_gradient", "make_divergence",
           "make_vector_gradient", "make_vector_laplace",
           "make_tensor_divergence", "make_poisson_solver"]
    


'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import numpy as np
import pytest

from .test_generic import iter_grids
from ..scalar import ScalarField
from ..base import FieldBase
from ...grids import UnitGrid, CartesianGrid, PolarGrid
from ...grids.boundaries import DomainError
from ...grids.tests.test_cartesian import _get_cartesian_grid
from ...tools.misc import module_available, skipUnlessModule



def test_interpolation_singular():
    """ test interpolation on singular dimensions """
    grid = UnitGrid([1])
    field = ScalarField(grid, data=3)
    
    # test constant boundary conditions
    bc = [{'type': 'value', 'value': 1}, {'type': 'value', 'value': 5}]
    x = np.linspace(0, 1, 7).reshape((7, 1))
    y = field.interpolate(x, method='numba', bc=bc)
    np.testing.assert_allclose(y, 1 + 4*x.ravel())

    # test derivative boundary conditions
    bc = [{'type': 'derivative', 'value': -2},
          {'type': 'derivative', 'value': 2}]
    x = np.linspace(0, 1, 7).reshape((7, 1))
    y = field.interpolate(x, method='numba', bc=bc)
    np.testing.assert_allclose(y, 2 + 2 * x.ravel())
    
    # test boundary interpolation
    for upper in [True, False]:
        val = field.get_boundary_values(axis=0, upper=upper, bc=[{'value': 1}])
        assert val == pytest.approx(1)



@pytest.mark.parametrize("grid", iter_grids())
def test_simple_shapes(grid):
    """ test simple scalar fields """
    pf = ScalarField.random_uniform(grid)
    np.testing.assert_equal(pf.data.shape, grid.shape)
    pf_lap = pf.laplace('natural')
    np.testing.assert_equal(pf_lap.data.shape, grid.shape)
    assert isinstance(pf.integral, float)
    
    pf_c = pf.copy()
    np.testing.assert_allclose(pf.data, pf_c.data)
    assert pf.grid == pf_c.grid
    assert pf.data is not pf_c.data
                
    if module_available("matplotlib"):
        pf.plot()  # simply test whether this does not cause errors


           
def test_scalars():
    """ test some scalar fields """
    grid = CartesianGrid([[0.1, 0.3], [-2, 3]], [3, 4])
    s1 = ScalarField(grid, np.full(grid.shape, 1))
    s2 = ScalarField(grid, np.full(grid.shape, 2))
    assert s1.average == pytest.approx(1)
    assert s1.magnitude == pytest.approx(1)
    
    s3 = s1 + s2
    assert s3.grid == grid
    np.testing.assert_allclose(s3.data, 3)
    s1 += s2
    np.testing.assert_allclose(s1.data, 3)
    
    s2 = FieldBase.from_state(s1.attributes, data=s1.data)
    assert s1 == s2
    assert s1.grid is s2.grid
    
    attrs = ScalarField.unserialize_attributes(s1.attributes_serialized)
    s2 = FieldBase.from_state(attrs, data=s1.data)
    assert s1 == s2
    assert s1.grid is not s2.grid
    
    # test options for plotting images
    if module_available("matplotlib"):
        s1.plot(transpose=True, colorbar=True)
        


def test_laplacian():
    """ test the gradient operator """
    grid = CartesianGrid([[0, 2 * np.pi], [0, 2 * np.pi]], [16, 16],
                         periodic=True)
    s = ScalarField.random_harmonic(grid, axis_combination=np.add, modes=1)

    s_lap = s.laplace('natural')
    assert s_lap.data.shape == (16, 16)
    np.testing.assert_allclose(s_lap.data, -s.data, rtol=0.1, atol=0.1)

    s.laplace('natural', out=s_lap)
    assert s_lap.data.shape == (16, 16)
    np.testing.assert_allclose(s_lap.data, -s.data, rtol=0.1, atol=0.1)



def test_gradient():
    """ test the gradient operator """
    grid = CartesianGrid([[0, 2 * np.pi], [0, 2 * np.pi]], [16, 16],
                         periodic=True)
    x, y = grid.cell_coords[..., 0], grid.cell_coords[..., 1]
    data = np.cos(x) + np.sin(y)
    
    s = ScalarField(grid, data)
    v = s.gradient('natural')
    assert v.data.shape == (2, 16, 16)
    np.testing.assert_allclose(v.data[0], -np.sin(x), rtol=0.1, atol=0.1)
    np.testing.assert_allclose(v.data[1], np.cos(y), rtol=0.1, atol=0.1)
   
    s.gradient('natural', out=v)
    assert v.data.shape == (2, 16, 16)
    np.testing.assert_allclose(v.data[0], -np.sin(x), rtol=0.1, atol=0.1)
    np.testing.assert_allclose(v.data[1], np.cos(y), rtol=0.1, atol=0.1)
      
        
        
@pytest.mark.parametrize("grid", iter_grids())
def test_interpolation_to_grid(grid):
    """ test whether data is interpolated correctly for different grids """
    sf = ScalarField.random_uniform(grid)
    sf2 = sf.interpolate_to_grid(grid, method='numba')
    np.testing.assert_allclose(sf.data, sf2.data, rtol=1e-6)


        
@pytest.mark.parametrize("grid", iter_grids())
def test_add_interpolated_scalar(grid):
    """ test the `add_interpolated` method """
    f = ScalarField(grid)
    a = np.random.random()
    
    c = tuple(grid.point_to_cell(grid.get_random_point()))
    p = grid.cell_to_point(c, cartesian=False)
    f.add_interpolated(p, a)
    assert f.data[c] == pytest.approx(a / grid.cell_volumes[c])
    
    f.add_interpolated(grid.get_random_point(cartesian=False), a)
    assert f.integral == pytest.approx(2 * a)
    
    f.data = 0  # reset
    add_interpolated = grid.make_add_interpolated_compiled()
    c = tuple(grid.point_to_cell(grid.get_random_point()))
    p = grid.cell_to_point(c, cartesian=False)
    add_interpolated(f.data, p, a)
    assert f.data[c] == pytest.approx(a / grid.cell_volumes[c])

    add_interpolated(f.data, grid.get_random_point(cartesian=False), a)
    assert f.integral == pytest.approx(2 * a)



def test_add_interpolated_1d():
    """ test the `add_interpolated` method for 1d systems """
    grid = PolarGrid(3, 5)
    f = ScalarField(grid)
    g = f.copy()
    a = np.random.random()
    for r in np.linspace(0, 3, 8).reshape(8, 1):
        f.data = g.data = 0
        f.add_interpolated(r, a)
        assert f.integral == pytest.approx(a)
        grid.make_add_interpolated_compiled()(g.data, r, a)
        np.testing.assert_array_almost_equal(f.data, g.data)

        
        
def test_random_harmonic():
    """ test whether random harmonic fields behave correctly """
    grid = _get_cartesian_grid(2)  # get random Cartesian grid
    x = ScalarField.random_harmonic(grid, modes=1)
    scaling = sum((2 * np.pi / L)**2 for L in grid.cuboid.size)
    y = -x.laplace('natural') / scaling
    np.testing.assert_allclose(x.data, y.data, rtol=1e-2, atol=1e-2)
    


def test_get_line_data():
    """ test different extraction methods for line data """
    grid = UnitGrid([16, 32])
    c = ScalarField.random_harmonic(grid)
    
    np.testing.assert_equal(c.get_line_data(extract='cut_0'),
                            c.get_line_data(extract='cut_x'))
    np.testing.assert_equal(c.get_line_data(extract='cut_1'),
                            c.get_line_data(extract='cut_y'))
    
    for extract in ['project_0', 'project_1', 'project_x', 'project_y']:
        data = c.get_line_data(extract=extract)['data_y']
        np.testing.assert_allclose(data, 0, atol=1e-14)
        


def test_from_expression():
    """ test creating scalar field from expression """
    grid = UnitGrid([1, 2])
    sf = ScalarField.from_expression(grid, "x * y")
    np.testing.assert_allclose(sf.data, [[0.25, 0.75]])
    
    
    
def test_interpolation_inhomogeneous_bc():
    """ test field interpolation with inhomogeneous boundary condition """
    sf = ScalarField(UnitGrid([3, 3], periodic=False))
    x = 1 + np.random.random()
    v = sf.interpolate([x, 0], method='numba',
                       bc=['natural', {'type': 'value', 'value': 'x'}])
    assert x == pytest.approx(v)



@skipUnlessModule("matplotlib")
def test_from_image(tmp_path):
    from matplotlib.pyplot import imsave
    img_data = np.random.uniform(size=(9, 8, 3))
    img_data_gray = img_data @ np.array([0.299, 0.587, 0.114])
    path = tmp_path / 'test_from_image.png'
    imsave(path, img_data, vmin=0, vmax=1)
    sf = ScalarField.from_image(path)
    np.testing.assert_allclose(sf.data, img_data_gray.T[:, ::-1], atol=0.05)
    


def test_to_scalar():
    """ test conversion to scalar field """
    sf = ScalarField.random_uniform(UnitGrid([3, 3]))
    np.testing.assert_allclose(sf.to_scalar().data, sf.data)    
    np.testing.assert_allclose(sf.to_scalar('squared_sum').data, sf.data**2)

    with pytest.raises(ValueError):
        sf.to_scalar('nonsense')



@pytest.mark.parametrize('grid', (g for g in iter_grids() if g.num_axes > 1))
@pytest.mark.parametrize('method', ['integral', 'average'])
def test_projection(grid, method):
    """ test scalar projection """
    sf = ScalarField.random_uniform(grid)
    for ax in grid.axes:
        sp = sf.project(ax, method=method)
        assert sp.grid.dim < grid.dim
        assert sp.grid.num_axes == grid.num_axes - 1
        if method == 'integral':
            assert sp.integral == pytest.approx(sf.integral)
        elif method == 'average':
            assert sp.average == pytest.approx(sf.average)
            
    with pytest.raises(ValueError):
        sf.project(grid.axes[0], method='nonsense')



@pytest.mark.parametrize('grid', (g for g in iter_grids() if g.num_axes > 1))
def test_slice(grid):
    """ test scalar slicing """
    sf = ScalarField(grid, 0.5)
    p = grid.get_random_point()
    for i in range(grid.num_axes):
        sf_slc = sf.slice({grid.axes[i]: p[i]})
        np.testing.assert_allclose(sf_slc.data, 0.5)
        assert sf_slc.grid.dim < grid.dim
        assert sf_slc.grid.num_axes == grid.num_axes - 1
        
    with pytest.raises(DomainError):
        sf.slice({grid.axes[0]: -10})
    with pytest.raises(ValueError):
        sf.slice({'q': 0})



def test_slice_positions():
    """ test scalar slicing at standard positions """
    grid = UnitGrid([3, 1])
    sf = ScalarField(grid, np.arange(3).reshape(3, 1))
    assert sf.slice({'x': 'min'}).data == 0
    assert sf.slice({'x': 'mid'}).data == 1
    assert sf.slice({'x': 'max'}).data == 2

    with pytest.raises(ValueError):
        sf.slice({'x': 'foo'})
    with pytest.raises(ValueError):
        sf.slice({'x': 0}, method='nonsense')



@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_poisson_solver_1d():
    """ test the poisson solver on 1d grids """
    # solve Laplace's equation
    grid = UnitGrid([4])
    field = ScalarField(grid)
    res = field.solve_poisson([{'value': -1}, {'value': 3}])
    np.testing.assert_allclose(res.data, grid.axes_coords[0] - 1)
    
    res = field.solve_poisson([{'value': -1}, {'derivative': 1}])
    np.testing.assert_allclose(res.data, grid.axes_coords[0] - 1)

    # test Poisson equation with 2nd Order BC
    res = field.solve_poisson([{'value': -1}, 'extrapolate'])
    
    # solve Poisson's equation
    grid = CartesianGrid([[0, 1]], 4)
    field = ScalarField(grid, data=1)
    
    res = field.copy()
    field.solve_poisson([{'value': 1}, {'derivative': 1}], out=res)
    xs = grid.axes_coords[0]
    np.testing.assert_allclose(res.data, 1 + 0.5 * xs**2, rtol=1e-2)
    
    # test inconsistent problem
    field.data = 1
    with pytest.raises(RuntimeError, match='Neumann'):
        field.solve_poisson({'derivative': 0})
    


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_poisson_solver_2d():
    """ test the poisson solver on 2d grids """
    grid = CartesianGrid([[0, 2 * np.pi]] * 2, 16)
    bcs = [{'value': 'sin(y)'}, {'value': 'sin(x)'}]
    
    # solve Laplace's equation
    field = ScalarField(grid)
    res = field.solve_poisson(bcs)
    xs = grid.cell_coords[..., 0]
    ys = grid.cell_coords[..., 1]
    
    # analytical solution was obtained with Mathematica
    expect = (np.cosh(np.pi - ys) * np.sin(xs) +
              np.cosh(np.pi - xs) * np.sin(ys)) / np.cosh(np.pi)
    np.testing.assert_allclose(res.data, expect, atol=1e-2, rtol=1e-2)
    
    # test more complex case for exceptions
    res = field.solve_poisson([{'value': 'sin(y)'}, {'curvature': 'sin(x)'}])
    
    
    
def test_interpolation_mutable():
    """ test interpolation on mutable fields """
    grid = UnitGrid([2], periodic=True)
    field = ScalarField(grid)
    
    for method in ['numba', 'scipy']:
        field.data = 1
        np.testing.assert_allclose(field.interpolate([0.5], method=method), 1)
        field.data = 2
        np.testing.assert_allclose(field.interpolate([0.5], method=method), 2)
        
    # test overwriting field values
    data = np.full_like(field.data, 3)
    intp = field.make_interpolator(method='numba')
    np.testing.assert_allclose(intp(np.array([0.5]), data), 3)



def test_boundary_interpolation_1d():
    """ test boundary interpolation for 1d fields """
    grid = UnitGrid([5])
    field = ScalarField(grid, np.arange(grid.shape[0]))
    
    # test boundary interpolation
    bndry_val = 0.25
    for bndry in grid._iter_boundaries():
        val = field.get_boundary_values(*bndry, bc={'value': bndry_val})
        np.testing.assert_allclose(val, bndry_val)
    
        ev = field.make_get_boundary_values(*bndry, bc={'value': bndry_val})
        out = ev()
        np.testing.assert_allclose(out, bndry_val)
        ev(data=field.data, out=out)
        np.testing.assert_allclose(out, bndry_val)
    


def test_boundary_interpolation_2d():
    """ test boundary interpolation for 2d fields """
    grid = CartesianGrid([[0.1, 0.3], [-2, 3]], [3, 3])
    field = ScalarField.random_normal(grid)
    
    # test boundary interpolation
    bndry_val = np.random.randn(3)
    for bndry in grid._iter_boundaries():
        val = field.get_boundary_values(*bndry, bc={'value': bndry_val})
        np.testing.assert_allclose(val, bndry_val)
    
        ev = field.make_get_boundary_values(*bndry, bc={'value': bndry_val})
        out = ev()
        np.testing.assert_allclose(out, bndry_val)
        ev(data=field.data, out=out)
        np.testing.assert_allclose(out, bndry_val)
    


'''
Solvers for Poisson's and Laplace's equation

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

from ..grids.base import GridBase
from ..grids.boundaries.axes import BoundariesData  # @UnusedImport
from ..fields import ScalarField
from ..tools.docstrings import fill_in_docstring



@fill_in_docstring
def solve_poisson_equation(rhs: ScalarField,
                           bc: "BoundariesData",
                           label: str = "Solution to Poisson's equation") \
                                -> ScalarField:
    r""" Solve Laplace's equation on a given grid
         
    Denoting the current field by :math:`u`, we thus solve for :math:`f`,
    defined by the equation 
    
    .. math::
        \nabla^2 u(\boldsymbol r) = -f(\boldsymbol r)
        
    with boundary conditions specified by `bc`.
        
    Note:
        In case of periodic or Neumann boundary conditions, the right hand
        side :math:`f(\boldsymbol r)` needs to satisfy the following
        condition
        
        .. math::
            \int f \, \mathrm{d}V = \oint g \, \mathrm{d}S \;,
            
        where :math:`g` denotes the function specifying the outwards
        derivative for Neumann conditions. Note that for periodic boundaries
        :math:`g` vanishes, so that this condition implies that the integral 
        over
        :math:`f` must vanish for neutral Neumann or periodic conditions.
    
    Args:
        rhs (:class:`~pde.fields.scalar.ScalarField`):
            The scalar field :math:`f` describing the right hand side
        bc:
            The boundary conditions applied to the field.
            {ARG_BOUNDARIES}
        label (str):
            The label of the returned field.
            
    Returns:
        :class:`~pde.fields.scalar.ScalarField`: The field :math:`u` that solves
        the equation. This field will be defined on the same grid as `rhs`.
    """
    # solve the poisson problem
    solve_poisson = rhs.grid.get_operator('poisson_solver', bc=bc)
    try:
        result = solve_poisson(rhs.data)
    except RuntimeError:
        average = rhs.average
        if abs(average) > 1e-10:
            raise RuntimeError('Could not solve the Poisson problem. One '
                               'possible reason for this is that only '
                               'periodic or Neumann conditions are '
                               'applied although the average of the field '
                               f'is {average} and thus non-zero.')
        else:
            raise  # another error occurred
     
    return ScalarField(rhs.grid, result, label=label)



@fill_in_docstring
def solve_laplace_equation(grid: GridBase,
                           bc: "BoundariesData",
                           label: str = "Solution to Laplace's equation") \
                                -> ScalarField:
    """ Solve Laplace's equation on a given grid.
    
    This is implemented by calling :func:`solve_poisson_equation` with a 
    vanishing right hand side.
    
    Args:
        grid (:class:`~pde.grids.base.GridBase`):
            The grid on which the equation is solved
        bc:
            The boundary conditions applied to the field.
            {ARG_BOUNDARIES}
        label (str):
            The label of the returned field.
            
    Returns:
        :class:`~pde.fields.scalar.ScalarField`: The field that solves the
        equation. This field will be defined on the given `grid`.
    """
    return solve_poisson_equation(ScalarField(grid), bc=bc, label=label)



"""
Custom PDE in one dimension
===========================

This example implements a PDE that is only defined in one dimension.
Here, we chose the `Korteweg-de Vries equation
<https://en.wikipedia.org/wiki/Korteweg–de_Vries_equation>`_, given by

.. math::
    \partial_t \phi = 6 \phi \partial_x \phi - \partial_x^2 \phi
    
which we implement using a custom PDE class below.
"""

from math import pi

from pde import (CartesianGrid, ScalarField, PDEBase, MemoryStorage,
                 plot_kymograph)


class KortewegDeVriesPDE(PDEBase):
    """ Korteweg-de Vries equation """
    
    def evolution_rate(self, state, t=0):
        """ implement the python version of the evolution equation """
        assert state.grid.dim == 1  # ensure the state is one-dimensional
        grad = state.gradient('natural')[0]
        return 6 * state * grad - grad.laplace('natural')


# initialize the equation and the space
grid = CartesianGrid([[0, 2*pi]], [32], periodic=True)
state = ScalarField.from_expression(grid, "sin(x)")

# solve the equation and store the trajectory
storage = MemoryStorage()
eq = KortewegDeVriesPDE()             
eq.solve(state, t_range=3, tracker=storage.tracker(.1))

# plot the trajectory as a space-time plot
plot_kymograph(storage)
